theory Lustre
imports Lustre_consistency
begin


section{* Basic Lustre *}


subsection{* Synchronous operators *}


lift_definition pre:: "'a::null stream \<Rightarrow> 'a stream" is
"\<lambda> S n. if n = 0 then null else (S (n - 1))" .

lemma pre_base[simp] : "(pre S)\<lbrakk>0\<rbrakk> = null"
by(simp add: pre_def sel_def Abs_stream_inverse')

lemma pre_step[simp] : "n > 0 \<Longrightarrow> (pre S)\<lbrakk>n\<rbrakk> = S\<lbrakk>n-1\<rbrakk>"
by(simp add: pre_def sel_def Abs_stream_inverse')


lemma pre_simp[simp, code_abbrev]: 
"(pre A)\<lbrakk>n\<rbrakk> = (if n = 0 then null else (A\<lbrakk>n - 1\<rbrakk>))"
by (auto simp: pre_def sel_def Abs_stream_inverse[simplified])


lift_definition followed_by:: "'a::null stream \<Rightarrow> 'a stream \<Rightarrow> 'a stream" (infixl "fby" 60) is
"\<lambda> S S' n. if n = 0 then (S 0) else (S' n)"  .

notation followed_by  (infixr "->" 25)

lemma fby_simp[simp, code_abbrev]: 
"(A -> B)\<lbrakk>n\<rbrakk> = (if n = 0 then A\<lbrakk>0\<rbrakk> else B\<lbrakk>n\<rbrakk>)"
by (auto simp: sel_def followed_by_def Abs_stream_inverse[simplified])

lemma fby_base[simp] : "(A -> B)\<lbrakk>0\<rbrakk> = A\<lbrakk>0\<rbrakk> "
by(simp)

lemma fby_step[simp] : "n > 0 \<Longrightarrow> (A -> B)\<lbrakk>n\<rbrakk> = B\<lbrakk>n\<rbrakk>"
by(simp)


subsection{* Generalized Follow. *}
(* Really useful ? *)

lift_definition gfb:: "'a::null stream \<Rightarrow> nat \<Rightarrow>'a stream \<Rightarrow> 'a stream" ("_  \<langle>:_:\<rangle> _" 60) 
is    "\<lambda> S k S' n. if n \<le> k then (S n) else (S' n)"  .

lemma gfb_simp[simp, code_abbrev]: 
"(A \<langle>:k:\<rangle> B)\<lbrakk>n\<rbrakk> = (if n \<le> k then A\<lbrakk>n\<rbrakk> else B\<lbrakk>n\<rbrakk>)"
by (auto simp: sel_def gfb_def Abs_stream_inverse[simplified])

lemma gfb_le[simp] : "n \<le> k \<Longrightarrow> (A \<langle>:k:\<rangle> B)\<lbrakk>n\<rbrakk> = A\<lbrakk>n\<rbrakk> "
by(simp)

lemma gfb_gr[simp] : "n > k \<Longrightarrow> (A \<langle>:k:\<rangle> B)\<lbrakk>n\<rbrakk> = B\<lbrakk>n\<rbrakk>"
by(simp) 

lemma gfp_vs_fp : "(A \<langle>:0:\<rangle> B) = (A -> B)"
apply(simp add: gfb_def followed_by_def)
apply(rule_tac f=Abs_stream in arg_cong, rule ext)
apply(auto)
done






subsection{* Constructor for Streams *}

definition mkstream :: "'a list \<Rightarrow> 'a option stream" 
where     "mkstream S \<equiv> Abs_stream (\<lambda>n. if n < length S 
                                        then Some (S!n) 
                                        else None)" 

notation mkstream ("\<lbrace>_\<rbrace>" 100)

thm mkstream_def

lemma [simp,code_unfold] : "(\<lbrace>S\<rbrace>\<lbrakk>n\<rbrakk>) = (if n < length S then Some (S ! n) else None)"
by(simp add: mkstream_def sel_def Abs_stream_inverse[simplified])


value "\<lbrace>[1::int,2,3]\<rbrace>\<lbrakk>2\<rbrakk>"  (* This should work ! *)
lemma "\<lbrace>[1::int,2,3]\<rbrace>\<lbrakk>2\<rbrakk> = Some 3"
by(simp)

subsection{* General schemes for strict operators *}

abbreviation
"strict_unary uop \<equiv> \<lambda> x. case x of None \<Rightarrow> None 
                                 | Some xx \<Rightarrow> Some (uop xx)"

definition strict_unary':: "('a::null \<Rightarrow> 'b::null) \<Rightarrow> 'a \<Rightarrow> 'b"
where "strict_unary' uop \<equiv> \<lambda> x. if is_def x then uop x else null"
(* 
lemma cc_unary_preserve : "\<langle>P\<rangle>\<^sub>\<lbrakk>\<^bsub>n\<^esub>\<^sub>\<rbrakk> \<Longrightarrow> \<langle>\<lambda>X. strict_unary' uop ((P X) \<lbrakk>n\<rbrakk>)\<rangle>\<^sub>\<lbrakk>\<^bsub>n\<^esub>\<^sub>\<rbrakk> "
*)
abbreviation 
"strict_binary bop \<equiv> \<lambda> x y. case x of None \<Rightarrow> None 
                                    | Some xx \<Rightarrow> case y of None \<Rightarrow> None 
                                                         | Some yy \<Rightarrow> Some (bop xx yy)"

lemma eq_upto_surj_pairing : 
"S \<triangleq>\<^sub>\<lbrakk>\<^bsub>n\<^esub>\<^sub>\<rbrakk> T \<Longrightarrow>  (fst\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m S \<triangleq>\<^sub>\<lbrakk>\<^bsub>n\<^esub>\<^sub>\<rbrakk> fst\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m T) \<and> (snd\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m S \<triangleq>\<^sub>\<lbrakk>\<^bsub>n\<^esub>\<^sub>\<rbrakk> snd\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m T)"
by (simp add: equal_upto_def)




lemma strict_binary_commute:
"comp_fun_commute (f) \<Longrightarrow> comp_fun_commute (strict_binary (f))"
by (auto simp: comp_fun_commute_def o_def fun_eq_iff split: option.split)


instantiation option :: (zero) zero
begin
lift_definition zero_option :: "'a option"  is "Some 0" .

instance ..
end


instantiation stream :: (zero) zero
begin
lift_definition zero_stream :: "'a stream"  is "\<lambda>n. 0" .
instance ..
end




lemma zero_simp[simp, code_abbrev]: "0\<lbrakk>n\<rbrakk> = Some 0"
by (auto simp: sel_def zero_stream_def zero_option_def Abs_stream_inverse[simplified])

lemma  const_stream_zero[simp]: "const_stream 0"
by(auto simp: const_stream_def zero_stream_def sel_def Abs_stream_inverse')


instantiation option :: (one) one
begin
lift_definition one_option :: "'a option" is "Some 1" .
instance ..
end


instantiation stream :: (one) one
begin
lift_definition one_stream :: "'a::one stream" is "\<lambda>n. 1" .
instance ..
end


text{* Note that this slight generalization allows also real-streams to
have 1's ... *}

lemma one_simp[simp, code_abbrev]: "1\<lbrakk>n\<rbrakk>  = Some 1"
by (auto simp: one_stream_def one_option_def sel_def Abs_stream_inverse')

lemma  const_stream_1[simp]: "const_stream 1"
by(auto simp: const_stream_def one_stream_def sel_def Abs_stream_inverse')



instantiation option :: (numeral) numeral
begin
definition plus_option :: "'a option \<Rightarrow> 'a option \<Rightarrow> 'a option" where
"x + y \<equiv> strict_binary (op +) x y"
instance proof
  fix a b c :: "'a option"
  show "a + b + c = a + (b + c)"
  by (auto simp: plus_option_def fun_eq_iff add.assoc split: option.split)
  qed
end

instantiation stream :: (numeral) numeral
begin
lift_definition plus_stream :: "'a stream \<Rightarrow> 'a stream \<Rightarrow> 'a stream" is
"\<lambda> x y n. (x n) + (y n)" .
instance proof
  fix a b c :: "'a stream"
  show "a + b + c = a + (b + c)"
  by (auto simp: fun_eq_iff plus_stream_def sel_def Abs_stream_inverse Abs_stream_inject add.assoc)
  qed
end

lemma plus_simp[simp, code_abbrev]: 
 "((A::'a::numeral option stream) + B)\<lbrakk>n\<rbrakk> = strict_binary (op +) (A\<lbrakk>n\<rbrakk>) (B\<lbrakk>n\<rbrakk>)"
by (auto simp: sel_def plus_stream_def plus_option_def Abs_stream_inverse split: option.split)


interpretation plus_profile : bin_scheme_option "(\<lambda>(X::('a::numeral) option stream) Y. X + Y)" "(op +)"
apply unfold_locales
apply (subst stream_ext)
apply (auto simp: plus_stream_def sel_def Abs_stream_inverse' plus_option_def)
by (simp add: option.case_eq_if)



instantiation option :: (minus) minus
begin
definition minus_option :: "'a option \<Rightarrow> 'a option \<Rightarrow> 'a option" where
"x - y \<equiv> strict_binary (op -) x y"
instance ..
end

instantiation stream :: (minus) minus
begin
lift_definition minus_stream :: "'a stream \<Rightarrow> 'a stream \<Rightarrow> 'a stream" 
  is "\<lambda>x y n. (x n) - (y n)" .
instance ..
end


interpretation minus_profile : bin_scheme_option "(\<lambda>(X::('a::minus) option stream) Y. X - Y)" "op -"
apply unfold_locales
apply(subst stream_ext)
apply(auto simp: minus_stream_def sel_def Abs_stream_inverse' minus_option_def)
by (simp add: option.case_eq_if)

text{* This gives us a bunch of lemmas: namely 14 for the moment. We need all these lemmas
       for all operators...*}
find_theorems name:minus_profile



instantiation option :: (uminus) uminus
begin
definition uminus_option :: "'a option \<Rightarrow> 'a option" where
"- x = strict_unary uminus x"
instance ..
end

instantiation stream :: (uminus) uminus
begin
lift_definition uminus_stream :: "'a stream \<Rightarrow> 'a stream" 
  is "\<lambda>x n. - (x n)" .
instance ..
end

lemma uminus_simp[simp, code_abbrev]: 
"(- A)\<lbrakk>n\<rbrakk> = strict_unary uminus (A\<lbrakk>n\<rbrakk>)"
by (auto simp: sel_def uminus_stream_def uminus_option_def Abs_stream_inverse)

interpretation uminus_profile : un_scheme_option "(\<lambda>(X::('a::uminus) option stream). - X)" "uminus"
apply unfold_locales
apply(subst stream_ext)
by(auto simp: uminus_stream_def sel_def Abs_stream_inverse' uminus_option_def)


instantiation option :: (times) times
begin
definition times_option :: "'a option \<Rightarrow> 'a option \<Rightarrow> 'a option" where
"x * y = strict_binary (op *) x y"
instance ..
end

instantiation stream :: (times) times
begin
lift_definition times_stream :: "'a stream \<Rightarrow> 'a stream \<Rightarrow> 'a stream" 
  is "\<lambda>x y n. (x n) * (y n)" .
instance ..
end

lemma times_simp[simp, code_abbrev]: 
"(A * B)\<lbrakk>n\<rbrakk> = strict_binary (op *) (A\<lbrakk>n\<rbrakk>) (B\<lbrakk>n\<rbrakk>)"
by (auto simp: sel_def times_stream_def times_option_def Abs_stream_inverse 
         split: option.split)


interpretation times_profile : bin_scheme_option "(\<lambda>(X::('a::times) option stream) Y. X * Y)" "op *"
apply unfold_locales
apply(subst stream_ext)
apply(auto simp: times_stream_def sel_def Abs_stream_inverse' times_option_def)
by (simp add: option.case_eq_if)



instantiation option :: (Divides.div) Divides.div
begin
definition div_option :: "'a option \<Rightarrow> 'a option \<Rightarrow> 'a option" where
"x div y = strict_binary (op div) x y"
definition mod_option :: "'a option \<Rightarrow> 'a option \<Rightarrow> 'a option" where
"x mod y = strict_binary (op mod) x y"
instance ..
end

instantiation stream :: (Divides.div) Divides.div
begin
lift_definition div_stream :: "'a stream \<Rightarrow> 'a stream \<Rightarrow> 'a stream" 
  is "\<lambda>x y n. (x n) div (y n)" .
lift_definition mod_stream :: "'a stream \<Rightarrow> 'a stream \<Rightarrow> 'a stream" 
  is "\<lambda>x y n. (x n) mod (y n)" .
instance ..
end

lemma div_simp[simp, code_abbrev]: 
"(A div B)\<lbrakk>n\<rbrakk> = strict_binary (op div) (A\<lbrakk>n\<rbrakk>) (B\<lbrakk>n\<rbrakk>)"
by (auto simp: sel_def div_stream_def div_option_def Abs_stream_inverse[simplified] 
         split: option.split)


lemma mod_simp[simp, code_abbrev]: 
"(A mod B)\<lbrakk>n\<rbrakk> = strict_binary (op mod) (A\<lbrakk>n\<rbrakk>) (B\<lbrakk>n\<rbrakk>)"
by (auto simp: sel_def mod_stream_def mod_option_def Abs_stream_inverse[simplified] 
         split: option.split)


interpretation div_profile : bin_scheme_option 
                             "(\<lambda>(X::('a::Divides.div) option stream) Y. X div Y)" "op div"
apply unfold_locales
apply(subst stream_ext)
apply(auto simp: div_stream_def sel_def Abs_stream_inverse' div_option_def)
by (simp add: option.case_eq_if)


interpretation mod_profile : bin_scheme_option 
                             "(\<lambda>(X::('a::Divides.div) option stream) Y. X mod Y)" "op mod"
apply unfold_locales
apply(subst stream_ext)
apply(auto simp: mod_stream_def sel_def Abs_stream_inverse' mod_option_def)
by (simp add: option.case_eq_if)


instantiation option :: (division_ring) inverse
begin
definition inverse_option :: "'a option \<Rightarrow> 'a option" where
"inverse x = strict_unary inverse x"
definition divide_option :: "'a option \<Rightarrow> 'a option \<Rightarrow> 'a option" where
"x / y = strict_binary (op /) x y"
instance ..
end

instantiation stream :: (inverse) inverse
begin
lift_definition inverse_stream :: "'a stream \<Rightarrow> 'a stream" 
  is "\<lambda>x n. inverse (x n)" .
lift_definition divide_stream :: "'a stream \<Rightarrow> 'a stream \<Rightarrow> 'a stream" 
  is "\<lambda>x y n. (x n) / (y n)" .
instance ..
end

lemma inverse_simp[simp, code_abbrev]: 
"(inverse A)\<lbrakk>n\<rbrakk> = strict_unary inverse (A\<lbrakk>n\<rbrakk>)"
by (auto simp: sel_def inverse_stream_def inverse_option_def Abs_stream_inverse[simplified])


lemma divide_simp[simp, code_abbrev]: 
"(A / B)\<lbrakk>n\<rbrakk> = strict_binary (op /) (A\<lbrakk>n\<rbrakk>) (B\<lbrakk>n\<rbrakk>)"
by (auto simp: sel_def divide_stream_def divide_option_def Abs_stream_inverse[simplified] split: option.split)


interpretation inverse_profile : un_scheme_option 
                                 "(\<lambda>(X::('a::division_ring) option stream). inverse X)" "inverse"
apply unfold_locales
apply(subst stream_ext)
by(auto simp: inverse_stream_def sel_def Abs_stream_inverse' inverse_option_def)


interpretation divide_profile : bin_scheme_option 
                                "(\<lambda>(X::('a::division_ring) option stream) Y. X / Y)" "op /"
apply unfold_locales
apply(subst stream_ext)
apply(auto simp: divide_stream_def sel_def Abs_stream_inverse' divide_option_def)
by (simp add: option.case_eq_if)



definition eq :: "'a::null stream \<Rightarrow> 'a stream \<Rightarrow> Bool stream" (infixl "=s" 55) 
where  "S =s T \<equiv> Abs_stream(\<lambda> n. if  S\<lbrakk>n\<rbrakk> = null \<or>  T\<lbrakk>n\<rbrakk> = null 
                                 then None
                                 else Some(S\<lbrakk>n\<rbrakk> = T\<lbrakk>n\<rbrakk>))" 
notation eq (infixl "\<doteq>" 55) 

(* voila *)
interpretation eq_profile : bin_scheme_option "\<lambda> (S::'a option stream) T. S =s T" "op ="
apply unfold_locales
apply(auto simp: scc_def eq_def  sel_def Abs_stream_inverse' )
by(subst stream_ext, auto simp: sel_def Abs_stream_inverse' minus_option_def)



lemma eq_simp[simp, code_abbrev]: 
"(A \<doteq> B)\<lbrakk>n\<rbrakk> = strict_binary (op =) (A\<lbrakk>n\<rbrakk>) (B\<lbrakk>n\<rbrakk>)"
by (auto simp: sel_def eq_def Abs_stream_inverse[simplified] split: option.split)


definition strong_eq :: "'a::null stream \<Rightarrow> 'a stream \<Rightarrow> Bool stream" (infixl "\<triangleq>" 55) 
where  "S \<triangleq> T \<equiv> Abs_stream(\<lambda> n.  Some(S\<lbrakk>n\<rbrakk> = T\<lbrakk>n\<rbrakk>))" 


subsection{* Not and some structured proofs *}

lift_definition not :: "Bool stream \<Rightarrow> Bool stream"  is
"\<lambda>x n. strict_unary Not (x n)" .

text{* heureka lemma: If a term-context @{term "P X"} is computationally safe, 
       then a super-term-context @{term "not(P X)"} is computationally safe.
       This also holds for strong computational consistency.
   *}

interpretation not_profile : un_scheme_option "not" "Not"
apply unfold_locales 
apply(subst stream_ext)
by(auto simp: not_def sel_def Abs_stream_inverse' divide_option_def)



text{* Now a proof based on unit-propagation, avoiding unfolding definitions and resulting
       in operational case-splitting. The resulting splitting-complexity is not depending on the size
       of the expression, just on the number of variables). This proof is deliberately slow-motion,
       in order to demonstrate the general principles. For a shorter version, see 
       the theorem \verb+and_commute+.*}
lemma not_idem : "not(not X) = X"
text{* Reducing to pointwise reasoning *}
proof(auto simp: stream_ext, rename_tac "n",case_tac "X\<lbrakk>n\<rbrakk>=null")
   fix        n::nat
   assume * :"X\<lbrakk>n\<rbrakk> = null"
   show      "(not (not X))\<lbrakk>n\<rbrakk> = X\<lbrakk>n\<rbrakk>"

             apply(insert *)
             text{* unit propagation *}
             apply(erule scc_subst2_rev)
             text{* establishing strong computational consistency for the two contexts *}
             apply(simp, simp)
             text{* and now just strictness evaluation (efficient deduction possible) *}
             by(simp)
next
   fix        n::nat
   assume * :"is_def (X\<lbrakk>n\<rbrakk>)"
   show       "(not (not X))\<lbrakk>n\<rbrakk> = X\<lbrakk>n\<rbrakk>"

             apply(insert *)
             text{* Reduction to HOL core: *}
             apply (simp  del :  Option.option.sel)
             text{* and now just computation of the boolean core (efficient deduction possible) *}
             apply(simp)
             (* actually, this separation is artificial and done for didactic reasons:
                by(simp add: not_nonstrict) does the trick entirely. *)
             done
qed

(* we need this schema of lemmas for all Lustre-operators including followed by (which is the 
only one requiring monotonicity of equality upto). This gives us a form of substitutivity and therefore
a means to exploit strictness principles. And the latter a means to avoid the splitting - blowup
that we get by simple unfolding. *)


lemma not_simp[simp, code_abbrev]: 
"(not A)\<lbrakk>n\<rbrakk> = strict_unary Not (A\<lbrakk>n\<rbrakk>)"
by (auto simp: sel_def not_def Abs_stream_inverse[simplified])

lift_definition "and" :: "Bool stream \<Rightarrow> Bool stream \<Rightarrow> Bool stream" ("_ and _") is
"\<lambda>x y n. strict_binary (op \<and>) (x n) (y n)" .

interpretation and_profile : bin_scheme_option "\<lambda> (S::Bool stream) T. S and T" "op \<and>"
apply unfold_locales
apply(subst stream_ext)
apply(auto simp: and_def sel_def Abs_stream_inverse' minus_option_def)
by (simp add: option.case_eq_if)

find_theorems name:and_profile

text{* Another abstract sample proof: *}
lemma and_commute: "(X and Y) = (Y and X)"
proof(auto simp: stream_ext, rename_tac "n",case_tac "X\<lbrakk>n\<rbrakk>=null")
   fix    n  assume * :"X\<lbrakk>n\<rbrakk> = null"    show  "X and Y\<lbrakk>n\<rbrakk> = Y and X\<lbrakk>n\<rbrakk>"
                                        by(rule scc_subst2_rev[OF *], simp_all) 
next 
   fix    n  assume **:"is_def (X\<lbrakk>n\<rbrakk>)"  show  "(X and Y)\<lbrakk>n\<rbrakk> = (Y and X)\<lbrakk>n\<rbrakk>"
      proof (case_tac "Y\<lbrakk>n\<rbrakk>=null")
             assume * :"Y\<lbrakk>n\<rbrakk> = null"    show  "(X and Y)\<lbrakk>n\<rbrakk> = (Y and X)\<lbrakk>n\<rbrakk>"
                                        by(rule scc_subst2_rev[OF *], simp_all)
      next   
             assume * :"is_def (Y\<lbrakk>n\<rbrakk>)"  show  "(X and Y)\<lbrakk>n\<rbrakk> = (Y and X)\<lbrakk>n\<rbrakk>"
                                        apply(insert * **)
                                        by(insert * **, auto)
      qed
qed

text{* This gives rise for a more general proof tactic for equal Lustre streams:
   \<bullet> apply extensionality, choose an arbitrary point n in the stream (pointwise argument)
   \<bullet> select the free stream variables or stream parameters of a goal
   \<bullet> make pointwise case destinction over them; exploit substitutivity and strictness rules
     for the cases that a free variable or parameter is  null in the point n.
   \<bullet> if all variables are defined, apply computational rules from the profile locale, 
     thus reduce the stream expression to a property inside plain HOL.

This leads to a \emph{linear} number of case-distinctions in the number of variables,
instead of an exponential number of distinctions in the size of the expressions.
*}


lift_definition or :: "Bool stream \<Rightarrow> Bool stream \<Rightarrow> Bool stream" ("_ or _") is
"\<lambda>x y n. strict_binary (op \<or>) (x n) (y n)" .

lemma or_simp[simp, code_abbrev]: 
"(A or B)\<lbrakk>n\<rbrakk> = strict_binary (op \<or>) (A\<lbrakk>n\<rbrakk>) (B\<lbrakk>n\<rbrakk>)"
by (auto simp: sel_def or_def Abs_stream_inverse[simplified] split: option.split)

interpretation or_profile : bin_scheme_option "\<lambda> (S::Bool stream) T. S or T" "op \<or>"
apply unfold_locales
apply(subst stream_ext)
apply(auto simp: or_def sel_def Abs_stream_inverse' minus_option_def)
by (simp add: option.case_eq_if)


lift_definition implies :: "Bool stream \<Rightarrow> Bool stream \<Rightarrow> Bool stream" ("_ => _") is
"\<lambda>x y n. strict_binary (op \<longrightarrow>) (x n) (y n)" .

lemma implies_simp[simp, code_abbrev]: 
"(A => B)\<lbrakk>n\<rbrakk> = strict_binary (op \<longrightarrow>) (A\<lbrakk>n\<rbrakk>) (B\<lbrakk>n\<rbrakk>)"
by (auto simp: sel_def implies_def Abs_stream_inverse[simplified] split: option.split)

interpretation implies_profile : bin_scheme_option "\<lambda> (S::Bool stream) T. S => T" "op \<longrightarrow>"
apply unfold_locales
apply(subst stream_ext)
apply(auto simp: implies_def sel_def Abs_stream_inverse' minus_option_def)
by (simp add: option.case_eq_if)



lift_definition xor :: "Bool stream \<Rightarrow> Bool stream \<Rightarrow> Bool stream" ("_ xor _") is
"\<lambda>x y n. strict_binary (op \<noteq>) (x n) (y n)" .

lemma xor_simp[simp, code_abbrev]: 
"(A xor B)\<lbrakk>n\<rbrakk> = strict_binary (op \<noteq>) (A\<lbrakk>n\<rbrakk>) (B\<lbrakk>n\<rbrakk>)"
by (auto simp: sel_def xor_def Abs_stream_inverse[simplified] split: option.split)


interpretation xor_profile : bin_scheme_option "\<lambda> (S::Bool stream) T. S xor T" "op \<noteq>"
apply unfold_locales
apply(subst stream_ext)
apply(auto simp: xor_def sel_def Abs_stream_inverse' minus_option_def)
by (simp add: option.case_eq_if)



(* problematic, better: use overloading instead of renaming <. Less syntactic ambiguities. *)
lift_definition lt :: "'a::{ord,null} option stream \<Rightarrow> 'a option stream \<Rightarrow> Bool stream" (" _ <s _") is
"\<lambda>x y n. strict_binary (op <) (x n) (y n)" .

interpretation lt_profile : bin_scheme_option "lt" "op <"
apply unfold_locales
apply(subst stream_ext)
apply(auto simp: lt_def sel_def Abs_stream_inverse' minus_option_def)
by (auto simp add: option.case_eq_if)

lemma lt_simp[simp, code_abbrev]: 
"(A <s B)\<lbrakk>n\<rbrakk> = (if  (A\<lbrakk>n\<rbrakk> = null) \<or>  (B\<lbrakk>n\<rbrakk> = null) then None else Some ((the (A\<lbrakk>n\<rbrakk>)) < (the (B\<lbrakk>n\<rbrakk>))))"
by (auto simp:  lt_def sel_def Abs_stream_inverse' option.case_eq_if)



(* 
At the long run, we must hide the internal split and close up the splitting redexes ...
declare lt_simp[simp del] 
*)

lift_definition le :: "'a::{ord,null} option stream \<Rightarrow> 'a option stream \<Rightarrow> Bool stream" (" _ <=s _") is
"\<lambda>x y n. strict_binary (op \<le>) (x n) (y n)" .

lemma le_simp[simp, code_abbrev]: 
"(A <=s B)\<lbrakk>n\<rbrakk> = (if  A\<lbrakk>n\<rbrakk> = null \<or>  B\<lbrakk>n\<rbrakk> = null then None else Some ((the (A\<lbrakk>n\<rbrakk>)) \<le> (the (B\<lbrakk>n\<rbrakk>))))"
by (auto simp:le_def sel_def Abs_stream_inverse' option.case_eq_if)

interpretation le_profile : bin_scheme_option "le" "op \<le>"
apply unfold_locales
apply(subst stream_ext)
apply(auto simp: le_def sel_def Abs_stream_inverse' minus_option_def)
by (auto simp add: option.case_eq_if)


(* @Abdou : turn that into a notation with commuted args ... ! ! !  *)
abbreviation gt (infixl ">s" 66) where "A >s B \<equiv> (B <s A)" 


abbreviation ge  (infixl ">=s" 67) where "A >=s B \<equiv> (B <=s A)"

abbreviation neq :: "'a::{ord,null} stream \<Rightarrow> 'a stream \<Rightarrow> Bool stream" (" _ <> _") 
where  "A<>B \<equiv> not(A \<doteq> B )" 

lemma neq_simp[simp, code_abbrev]: 
"(A <> B)\<lbrakk>n\<rbrakk> = (if  A\<lbrakk>n\<rbrakk> = null \<or>  B\<lbrakk>n\<rbrakk> = null then None else Some (A\<lbrakk>n\<rbrakk> \<noteq> B\<lbrakk>n\<rbrakk>))"
by (auto simp: sel_def not_def eq_def Abs_stream_inverse')






(* ... *)

lift_definition true::"Bool stream" is "(\<lambda>n. Some True)" .

lemma true_CS : "const_stream true"
by(auto simp: const_stream_def true_def sel_def Abs_stream_inverse')

lemma true_simp[simp, code_abbrev]: "true\<lbrakk>n\<rbrakk> = Some True"
by (auto simp: sel_def true_def Abs_stream_inverse')

lift_definition false::"Bool stream" is "(\<lambda>n. Some False)" .

lemma false_CS : "const_stream false"
by(auto simp: const_stream_def false_def sel_def Abs_stream_inverse')

lemma false_simp[simp, code_abbrev]: "false\<lbrakk>n\<rbrakk> = Some False"
by (auto simp: sel_def false_def Abs_stream_inverse[simplified])

subsection{* Conditional operators *}

definition If :: "Bool stream \<Rightarrow> 'a::null stream \<Rightarrow> 'a stream \<Rightarrow> 'a stream" 
                 ("(if (_)/ then (_)/ else (_) endif)" [0, 0, 10] 10)
where     "(if c then x else y endif) \<equiv> 
            Abs_stream(\<lambda>n. if c\<lbrakk>n\<rbrakk> = null then  null 
                           else if the(c\<lbrakk>n\<rbrakk>) then x\<lbrakk>n\<rbrakk> else y\<lbrakk>n\<rbrakk>)" 

lemma If_simp[simp, code_abbrev]: 
"(if c then P else Q endif)\<lbrakk>n\<rbrakk> =
  (case c\<lbrakk>n\<rbrakk> of None => None | 
               Some C => if C then P\<lbrakk>n\<rbrakk> else Q\<lbrakk>n\<rbrakk>)"
by (auto simp: sel_def If_def Abs_stream_inverse' split: option.split)


lemma [simp] : "(if true then P else Q endif) = P"
by(simp add: If_def true_def sel_def Rep_stream_inverse Abs_stream_inverse')

lemma [simp] : "(if false then P else Q endif) = Q"
by(simp add: If_def false_def sel_def Rep_stream_inverse Abs_stream_inverse')


definition assert'::"Bool stream \<Rightarrow> bool" 
where     "assert' b \<equiv> \<forall>n. \<not> (the (isNil b\<lbrakk>n\<rbrakk>))  \<and> (the (b\<lbrakk>n\<rbrakk>))"

definition assert::"Bool stream \<Rightarrow> bool" 
where     "assert b = (\<forall>n. case b\<lbrakk>n\<rbrakk> of None => False | Some X \<Rightarrow> X)"

lemma assert_sem : "assert b \<Longrightarrow> b\<lbrakk>n\<rbrakk> = (Some True)"
apply(simp add:assert_def)
apply(erule_tac x=n in allE)
by(auto split:option.split_asm)

(*really useful ? ? ? *)
lemma numeral_is_never_Nil[simp]: "\<not>the (isNil ((numeral N)::'a::numeral option stream)\<lbrakk>n\<rbrakk>)"
by (induct_tac N,
    auto simp: numeral_Bit0 numeral_Bit1 sel_def one_stream_def one_option_def 
               isNil_def plus_stream_def plus_option_def Abs_stream_inverse')

lemma assert_numeral_never_Nil:"assert (not (isNil ((numeral N)::'a::numeral option stream)))"
apply(auto simp:assert_def not_def)
by (induct_tac N,
    auto simp: numeral_Bit0 numeral_Bit1 sel_def one_stream_def one_option_def 
               isNil_def plus_stream_def plus_option_def Abs_stream_inverse')

(*really useful ? ? ? *)
lemma one_is_never_Nil[simp]: "\<not>the ((isNil (1::'a::numeral option stream))\<lbrakk>n\<rbrakk>)"
by (auto simp: sel_def isNil_def one_stream_def one_option_def null_option_def Abs_stream_inverse')

lemma zero_is_never_Nil[simp]: "\<not>the ((isNil (0::'a::zero option stream))\<lbrakk>n\<rbrakk>)"
by (auto simp: sel_def isNil_def zero_stream_def zero_option_def null_option_def Abs_stream_inverse')


lemma numeral_add_sel_distr[simp]: 
"\<not>the (isNil a\<lbrakk>n\<rbrakk>) \<Longrightarrow> \<not>the (isNil b\<lbrakk>n\<rbrakk>) \<Longrightarrow> ((a::'a::numeral option stream) + b)\<lbrakk>n\<rbrakk> = (a\<lbrakk>n\<rbrakk> + b\<lbrakk>n\<rbrakk>)"
by (auto simp: sel_def isNil_def plus_stream_def Abs_stream_inverse[simplified])


lemma numeral_add_one_sel_distr[simp]: 
"\<not>the (isNil a\<lbrakk>n\<rbrakk>) \<Longrightarrow> ((a::'a::numeral option stream) + 1)\<lbrakk>n\<rbrakk> = (a\<lbrakk>n\<rbrakk> + 1)"
by (subst numeral_add_sel_distr, 
    auto simp: sel_def isNil_def one_stream_def one_option_def null_option_def Abs_stream_inverse')

lemma numeral_add_one_nil_distr[simp]: 
"isNil ((a::'a::numeral option stream) + 1)\<lbrakk>n\<rbrakk> = isNil a\<lbrakk>n\<rbrakk>"
by (auto simp: sel_def isNil_def plus_stream_def null_option_def plus_option_def 
               one_stream_def one_option_def Abs_stream_inverse'
         split: option.split)

text{* A recursive block is characterized by the following principles:
\begin{enumerate}
\item they have a stream of internal states with type @{typ "'\<sigma>::null stream"}.
\item they have a global stream of inputs with type @{typ "'\<iota>::null stream"} and 
      a global stream of outputs with type @{typ "'o::null stream"}.
\item there is feed-back of the out-stream, \ie. the global stream of
      outputs becomes part of the input of the computation, represented
      by the \emph{body} of the recursive block.
\item the computation of a recursive block must be step-consistent :
      the computation of  its body can ONLY depend on previously computed
      points in input, state and output streams.
\end{enumerate}
*}


fun    loop\<^sub>0 :: "(('\<iota> stream \<times> ('\<sigma> stream \<times> 'o stream)) \<Rightarrow> ('\<sigma> stream \<times> 'o stream)) \<Rightarrow> 
                 '\<iota>::null stream \<Rightarrow> 
                 nat \<Rightarrow> 
                 ('\<sigma>::null stream \<times> 'o::null stream)"
where "loop\<^sub>0 body in\<^sub>a 0       = (let (\<sigma>, b) = body(in\<^sub>a, nil, nil) 
                                in (nil\<lbrakk>0:=\<sigma>\<lbrakk>0\<rbrakk>\<rbrakk>, nil\<lbrakk>0:=b\<lbrakk>0\<rbrakk>\<rbrakk>))"
     |"loop\<^sub>0 body in\<^sub>a (Suc n) = (let (\<sigma>, b) = loop\<^sub>0 body in\<^sub>a n;
                                    (\<sigma>', b') = body (in\<^sub>a, \<sigma>, b)
                                 in (\<sigma>\<lbrakk>Suc n:=\<sigma>'\<lbrakk>Suc n\<rbrakk>\<rbrakk>, b\<lbrakk>Suc n:=b'\<lbrakk>Suc n\<rbrakk>\<rbrakk>))"

text{* The following definition establishes the global view on a recursive block:
       the internal state is actually hidden, and the output re-packaged. 
*}

lift_definition loop ::"(('a stream \<times> '\<sigma>::null stream \<times> 'b stream) \<Rightarrow> ('\<sigma> stream \<times> 'b stream)) \<Rightarrow> 
                        'a::null stream \<Rightarrow>
                        'b::null stream"
is "\<lambda> body in\<^sub>a. (\<lambda>n. let (\<sigma>, b) = loop\<^sub>0 body in\<^sub>a n in (b\<lbrakk>n\<rbrakk>))" .

 
lemma loop_simp: "(loop b i)\<lbrakk>n\<rbrakk> = (snd (loop\<^sub>0 b i n))\<lbrakk>n\<rbrakk>"
by (auto simp: loop_def sel_def Abs_stream_inverse' split: Product_Type.split_split )


lemma loop_base[simp]: "(loop b i)\<lbrakk>0\<rbrakk> = (snd (b(i, nil, nil)))\<lbrakk>0\<rbrakk>"
by (auto simp: loop_def sel_def stream_upd_def Abs_stream_inverse' split: Product_Type.split_split)

lemma loop_step[simp]: "(loop b i)\<lbrakk>Suc n\<rbrakk> = (snd (b(i, loop\<^sub>0 b i n)))\<lbrakk>Suc n\<rbrakk>"
by (auto simp: sel_def stream_upd_simp[simplified sel_def] loop_def Abs_stream_inverse' split: Product_Type.split_split)


lemma stream_eq_iff: "(S = T) = (\<forall>n. S\<lbrakk>n\<rbrakk> = T\<lbrakk>n\<rbrakk>)"
by (auto simp: sel_def Rep_stream_inject[symmetric] fun_eq_iff)


lemma loop_suc_intro: 
"\<forall>st ou st' ou'. (st', ou') = (b(i, st, ou)) \<longrightarrow> P ((snd (b(i, st, ou)))\<lbrakk>Suc n\<rbrakk>) \<Longrightarrow> 
                 P ((loop b i)\<lbrakk>Suc n\<rbrakk>) "
apply (auto simp: loop_simp split: Product_Type.split_split)
apply (case_tac n, auto split: Product_Type.split_split Product_Type.split_split_asm)
apply (erule_tac x="nil\<lbrakk>0 := x1b\<lbrakk>0\<rbrakk>\<rbrakk>" in  allE)
apply (erule_tac x="nil\<lbrakk>0 := x2b\<lbrakk>0\<rbrakk>\<rbrakk>" in allE)
apply (auto)
apply (erule_tac x="x1b\<lbrakk>Suc nat := x1c\<lbrakk>Suc nat\<rbrakk>\<rbrakk>" in  allE)
apply (erule_tac x="x2b\<lbrakk>Suc nat := x2c\<lbrakk>Suc nat\<rbrakk>\<rbrakk>" in allE)
apply (auto)
done



lemma Abs_stream_sel[simp]: "(Abs_stream P)\<lbrakk>n\<rbrakk> = P n"
by (auto simp: sel_def Abs_stream_inverse')

lemma loop_induct:
assumes A: "P ((snd (b(i, nil, nil)))\<lbrakk>0\<rbrakk>)"
and    B: "\<And>n st ou st' ou'. (st', ou') = (b(i, st, ou)) \<Longrightarrow> P ((snd (b(i, st, ou)))\<lbrakk>n\<rbrakk>) \<Longrightarrow> 
                P ((snd (b(i, st\<lbrakk>n := st'\<lbrakk>n\<rbrakk>\<rbrakk>, ou\<lbrakk>n := ou'\<lbrakk>n\<rbrakk>\<rbrakk>)))\<lbrakk>Suc n\<rbrakk>)"
shows "P ((loop b i)\<lbrakk>n\<rbrakk>)"
apply (rule nat_induct, auto simp: A loop_step)
apply (auto simp: loop_def loop_step split: Product_Type.split_split_asm)
apply (case_tac n, auto)
apply (auto split: Product_Type.split_split_asm simp: B loop_step)
done


lemma numeral_stream_is_numeral[simp, code_unfold]: "(numeral N)\<lbrakk>n\<rbrakk> = numeral N"
apply (induct_tac N)
by (auto simp: numeral_Bit0 numeral_Bit1 sel_def plus_stream_def one_stream_def Abs_stream_inverse')

lemma numeral_option_is_some_numeral[simp]: "(numeral N) = Some (numeral N)"
apply (simp)
apply (induct_tac N)
apply (auto simp: numeral_Bit0 numeral_Bit1 one_option_def)
apply (subst numeral_add)+
apply (auto simp only: plus_stream_def plus_option_def split: option.split option.split_asm)
apply (subst (asm) numeral_add)+
apply (auto simp: plus_stream_def plus_option_def split: option.split option.split_asm)
apply (subst numeral_add)+
apply (subst (asm) numeral_add)+
apply (auto simp only: plus_stream_def plus_option_def split: option.split option.split_asm)
apply (auto)
done

lemma numeral_stream_rep_is_numeral[simp, code_unfold] : "(numeral N)\<lbrakk>n\<rbrakk> = Some (numeral N)"
by (simp)


lemma Rep_Abs_stream[simp]: "Rep_stream (Abs_stream (\<lambda> n. Some (y n))) na = Some (y na)"
by (auto simp: Abs_stream_inverse[simplified])


lemma sel_dest: "P (a\<lbrakk>n\<rbrakk>) \<Longrightarrow> (\<And> x. a\<lbrakk>n\<rbrakk> = Some x \<Longrightarrow> P (Some x))"
apply (auto simp: sel_def)
apply (case_tac "Rep_stream a n", auto)
done

lemma sel_intro: "\<And> x. a\<lbrakk>n\<rbrakk> = Some x \<Longrightarrow> P (Some x) \<Longrightarrow> P (a\<lbrakk>n\<rbrakk>)"
by (auto simp: sel_def)

declare Abs_stream_inverse'[simp] (* really ? *)

end

