theory Lustre_basics
imports Lustre_types
begin



text{* The constant null-stream and the the only non-strict test *}


lift_definition fst\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m :: "('a::null \<times> 'b::null) stream \<Rightarrow> 'a stream" is
"\<lambda> S n. fst(S n)" .

lift_definition snd\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m :: "('a::null \<times> 'b::null) stream \<Rightarrow> 'b stream" is
"\<lambda> S n. snd(S n)" .

term "(snd\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m (snd\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m (S::(Integer \<times> Bool \<times> Integer) stream)))
      :: Integer stream"



lift_definition Pair\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m::"'a::null stream \<Rightarrow> 'b::null stream \<Rightarrow> ('a \<times> 'b) stream " is
"\<lambda> a b n. (a n, b n)" .

definition split\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m :: "('a::null stream \<Rightarrow> 'b::null stream \<Rightarrow> 'c) \<Rightarrow> 
                         ('a \<times> 'b) stream \<Rightarrow> 'c"
where     "split\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m f S =  f (fst\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m S) (snd\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m S) "

term      "split\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m (\<lambda> (a::('a::null stream)) b::('b::null stream). P) 
                     (S::('a \<times> 'b) stream) "

term      "split\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m (\<lambda> (a::('a::null stream)).
                         split\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m (\<lambda>(b::('b::null stream))
                                      c::('c::null stream). P a b c)) 
                      (S::('a \<times> 'b \<times> 'c) stream)"

text{* On this basis, we can now rephrase common projection - repackaging machinery
       to streams.
 
       It is possible to define an analogous nested notation for 
       @{term "(\<lambda>(a,b,c). P a b c) S"}, 
       so something like:
       \inlineisar$(\<lambda>\<lbrakk>a,b,c\<rbrakk>. P a b c) S$
       on paired streams as for split on pairs.
*}



nonterminal stream_args and spatterns

syntax
  "_stream"      :: "'a stream => stream_args => ('a * 'b) stream"        ("(1'\<lbrakk>_,/ _'\<rbrakk>)")
  "_stream_arg"  :: "'a stream => stream_args"                   ("_")
  "_stream_args" :: "'a stream => stream_args => stream_args"     ("_,/ _")
  "_spattern"    :: "[pttrn, spatterns] => pttrn"         ("'\<lbrakk>_,/ _'\<rbrakk>")
  ""            :: "pttrn => spatterns"                  ("_")
  "_spatterns"   :: "[pttrn, spatterns] => spatterns"      ("_,/ _")



translations
  "\<lbrakk>x, y\<rbrakk>" == "CONST Pair\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m x y"
  "_spattern x y" => "CONST Pair\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m x y"
  "_spatterns x y" => "CONST Pair\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m x y"
  "_stream x (_stream_args y z)" == "_stream x (_stream_arg (_stream y z))"
  "%\<lbrakk>x, y, zs\<rbrakk>. b" == "CONST split\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m (%x \<lbrakk>y, zs\<rbrakk>. b)"
  "%\<lbrakk>x, y\<rbrakk>. b" == "CONST split\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m (%x y. b)"
  "_abs (CONST Pair\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m x y) t" => "%\<lbrakk>x, y\<rbrakk>. t"
  -- {* The last rule accommodates tuples in `case C ... (x,y) ... => ...'
     The (x,y) is parsed as `Pair x y' because it is logic, not pttrn *}

(*reconstruct pattern from (nested) splits, avoiding eta-contraction of body;
  works best with enclosing "let", if "let" does not avoid eta-contraction*)
print_translation {*
  let
    fun split_tr' [Abs (x, T, t as (Abs abs))] =
          (* split (%x y. t) => %(x,y) t *)
          let
            val (y, t') = Syntax_Trans.atomic_abs_tr' abs;
            val (x', t'') = Syntax_Trans.atomic_abs_tr' (x, T, t');
          in
            Syntax.const @{syntax_const "_abs"} $
              (Syntax.const @{syntax_const "_spattern"} $ x' $ y) $ t''
          end
      | split_tr' [Abs (x, T, (s as Const (@{const_syntax split\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m}, _) $ t))] =
          (* split (%x. (split (%y z. t))) => %(x,y,z). t *)
          let
            val Const (@{syntax_const "_abs"}, _) $
              (Const (@{syntax_const "_spattern"}, _) $ y $ z) $ t' = split_tr' [t];
            val (x', t'') = Syntax_Trans.atomic_abs_tr' (x, T, t');
          in
            Syntax.const @{syntax_const "_abs"} $
              (Syntax.const @{syntax_const "_spattern"} $ x' $
                (Syntax.const @{syntax_const "_spatterns"} $ y $ z)) $ t''
          end
      | split_tr' [Const (@{const_syntax split\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m}, _) $ t] =
          (* split (split (%x y z. t)) => %((x, y), z). t *)
          split_tr' [(split_tr' [t])] (* inner split_tr' creates next pattern *)
      | split_tr' [Const (@{syntax_const "_abs"}, _) $ x_y $ Abs abs] =
          (* split (%pttrn z. t) => %(pttrn,z). t *)
          let val (z, t) = Syntax_Trans.atomic_abs_tr' abs in
            Syntax.const @{syntax_const "_abs"} $
              (Syntax.const @{syntax_const "_spattern"} $ x_y $ z) $ t
          end
      | split_tr' _ = raise Match;
  in [(@{const_syntax split\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m}, K split_tr')] end
*}

(* print "split f" as "\<lambda>(x,y). f x y" and "split (\<lambda>x. f x)" as "\<lambda>(x,y). f x y" *) 
typed_print_translation {*
  let
    fun split_guess_names_tr' T [Abs (x, _, Abs _)] = raise Match
      | split_guess_names_tr' T [Abs (x, xT, t)] =
          (case (head_of t) of
            Const (@{const_syntax split\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m}, _) => raise Match
          | _ =>
            let 
              val (_ :: yT :: _) = binder_types (domain_type T) handle Bind => raise Match;
              val (y, t') = Syntax_Trans.atomic_abs_tr' ("y", yT, incr_boundvars 1 t $ Bound 0);
              val (x', t'') = Syntax_Trans.atomic_abs_tr' (x, xT, t');
            in
              Syntax.const @{syntax_const "_abs"} $
                (Syntax.const @{syntax_const "_spattern"} $ x' $ y) $ t''
            end)
      | split_guess_names_tr' T [t] =
          (case head_of t of
            Const (@{const_syntax split\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m}, _) => raise Match
          | _ =>
            let
              val (xT :: yT :: _) = binder_types (domain_type T) handle Bind => raise Match;
              val (y, t') =
                Syntax_Trans.atomic_abs_tr' ("y", yT, incr_boundvars 2 t $ Bound 1 $ Bound 0);
              val (x', t'') = Syntax_Trans.atomic_abs_tr' ("x", xT, t');
            in
              Syntax.const @{syntax_const "_abs"} $
                (Syntax.const @{syntax_const "_spattern"} $ x' $ y) $ t''
            end)
      | split_guess_names_tr' _ _ = raise Match;
  in [(@{const_syntax split\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m}, K split_guess_names_tr')] end
*}

term "(\<lambda>\<lbrakk>x, y\<rbrakk>. P x y)  :: ('a::null \<times> 'b::null) stream \<Rightarrow> 'c"

term "let \<lbrakk>x, y\<rbrakk> = a in b"

term "\<lbrakk>x, y\<rbrakk>"

(* This should do the projection business automatically and
   comfortably for arbitrarily structured streams ... *)

(* since records were mapped to pairs, this might work for
   Record notation too --- provided they were not abstracted via
   typedef *)

(* Thus, we can not generically describe an OPERATION with a type 
   for all tuples, but we can describe and support generically a TERM STRUCTURE
   that does the trick. *)



text{* The central destructor *}
definition      sel ::"'a stream \<Rightarrow> nat \<Rightarrow> 'a" ("_\<lbrakk>_\<rbrakk>" [9,0]500) 
where          "s\<lbrakk>n\<rbrakk> = (Rep_stream s n)"

lemma null_sel[simp] : "null\<lbrakk>n\<rbrakk> = null"
unfolding sel_def null_stream_def nil_def
by(auto simp: Abs_stream_inverse')


text{* The foundational theorem for pointwise reasoning on streams *}
lemma stream_ext : "(s = t) = (\<forall>n. s\<lbrakk>n\<rbrakk> = t\<lbrakk>n\<rbrakk>)"
apply(auto simp: sel_def)
apply(subst Rep_stream_inverse[symmetric])
apply(subst(3) Rep_stream_inverse[symmetric])
apply(rule_tac f=Abs_stream in arg_cong)
apply(rule ext) (* heureka ! *)
by auto

lemma nil_sel[simp]: "nil\<lbrakk>n\<rbrakk> = null"
by (auto simp: nil_def sel_def Rep_stream_inverse Abs_stream_inverse)

lemmas nil_simp = nil_sel (* legacy *)

lemma fst_stream_simp[simp]: "fst\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m \<lbrakk>x, y\<rbrakk> = x"
by (auto simp: fst\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m_def Pair\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m_def Rep_stream_inverse Abs_stream_inverse)

lemma snd_stream_simp[simp]: "snd\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m \<lbrakk>x, y\<rbrakk> = y"
by (auto simp: snd\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m_def Pair\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m_def Rep_stream_inverse Abs_stream_inverse)

lemma fst_stream_sel_simp[simp]: "(fst\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m P)\<lbrakk>n\<rbrakk> = fst(P\<lbrakk>n\<rbrakk>)"
by (auto simp: sel_def fst\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m_def Pair\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m_def Rep_stream_inverse Abs_stream_inverse)

lemma snd_stream_sel_simp[simp]: "(snd\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m P)\<lbrakk>n\<rbrakk> = snd(P\<lbrakk>n\<rbrakk>)"
by (auto simp: sel_def snd\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m_def Pair\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m_def Rep_stream_inverse Abs_stream_inverse)

lemma pair_stream_simp[simp]: "((\<lbrakk>a, b\<rbrakk>)\<lbrakk>n\<rbrakk>) = (a\<lbrakk>n\<rbrakk>, b\<lbrakk>n\<rbrakk>)"
by (auto simp: Pair\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m_def sel_def Rep_stream_inverse Abs_stream_inverse)

lemma stream_split_indep_simp[simp]: "(\<lambda>\<lbrakk>x, y\<rbrakk>. P) x = P"
by (simp add: split\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m_def)

lemma stream_split_indep1_simp[simp]: "(\<lambda>\<lbrakk>x, y\<rbrakk>. P x) \<lbrakk>x,y\<rbrakk> = P x"
by (simp add: split\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m_def)

lemma stream_split_indep2_simp[simp]: "(\<lambda>\<lbrakk>x, y\<rbrakk>. P y) \<lbrakk>x,y\<rbrakk> = P y"
by (simp add: split\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m_def)

lemma stream_split_simp[simp]: "((\<lambda>\<lbrakk>x, y\<rbrakk>. Q x y) a) = Q (fst\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m a) (snd\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m a)"
by (simp add: split\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m_def)

lemma stream_split_stream_pair_simp[simp]: "((\<lambda>\<lbrakk>x, y\<rbrakk>. Q x y)\<lbrakk>a, b\<rbrakk>) = Q a b"
by (simp add: split\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m_def)

lemma stream_split_pair_simp[simp]: 
"split\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m Pair (P \<lbrakk>a, b\<rbrakk>) = (fst\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m (P \<lbrakk>a, b\<rbrakk>), snd\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m (P \<lbrakk>a, b\<rbrakk>))"
by (auto simp: split\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m_def)

lemma stream_split_nil_simp[simp]: "(\<lambda>\<lbrakk>x, y\<rbrakk>. P x y) nil = P nil nil"
by (auto simp: split\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m_def nil_def null_prod_def fst\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m_def snd\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m_def Rep_stream_inverse Abs_stream_inverse)




text{* Stream updates *}

lift_definition stream_upd :: "'a::null stream \<Rightarrow> nat \<Rightarrow> 'a \<Rightarrow> 'a stream" is
"\<lambda> S n b m. if m=n then b else S m" .


nonterminal supdbinds and supdbind

syntax
  "_supdbind" :: "['a, 'a] => supdbind"               ("(2_ :=/ _)")
  ""          :: "supdbind => supdbinds"              ("_")
  "_supdbinds":: "[supdbind, supdbinds] => supdbinds" ("_,/ _")
  "_sUpdate"  :: "['a, supdbinds] => 'a"              ("_/'\<lbrakk>(_)'\<rbrakk>" [1000, 0] 900)

translations
  "_sUpdate f (_supdbinds b bs)" == "_sUpdate (_sUpdate f b) bs"
  "f\<lbrakk>x:=y\<rbrakk>" == "CONST stream_upd f x y"


lemma stream_upd_simp[simp] :
 "(f\<lbrakk>x:=y\<rbrakk>)\<lbrakk>n\<rbrakk> = (if x = n then y else f\<lbrakk>n\<rbrakk>)"
by (auto simp: sel_def stream_upd_def Abs_stream_inverse[simplified])

lemma stream_upd_id [simp]: "(f\<lbrakk>x:=y\<rbrakk>)\<lbrakk>x\<rbrakk> = y"
by(simp)

lemma stream_upd_nonid [simp]: "x\<noteq>z \<Longrightarrow> (f\<lbrakk>x:=y\<rbrakk>)\<lbrakk>z\<rbrakk> = f\<lbrakk>z\<rbrakk>"
by(simp)


lemma stream_upd_same[simp] : "(f\<lbrakk>x:=f\<lbrakk>x\<rbrakk>\<rbrakk>) = f"
by (auto simp: sel_def stream_upd_def Rep_stream_inject[symmetric] Abs_stream_inverse)


subsubsection{* A Rudimentary Code-Generator Setup *}


declare sel_def [code_unfold] 
        Rep_stream_inverse[code_unfold]
        Abs_stream_inverse'[code_unfold, code_abbrev]
                    
lemma [simp] : "((isNil (Abs_stream S)) \<lbrakk>n\<rbrakk>) \<noteq> None"
by(simp add: isNil_def sel_def Abs_stream_inverse[simplified] null_option_def)

lemma [simp] : "is_def ((isNil (Abs_stream S))\<lbrakk>n\<rbrakk>)"
by(simp add: isNil_def sel_def Abs_stream_inverse[simplified] null_option_def)
 

value  "(Abs_stream([0 \<mapsto> 2::int, 2 \<mapsto> 4, 4 \<mapsto> 7]))\<lbrakk>4\<rbrakk>"
value  "(isNil (Abs_stream [0 \<mapsto> 2::int, 2 \<mapsto> 4, 4 \<mapsto> 7])) \<lbrakk>3\<rbrakk> " (* yuck ! ! ! *)
lemma  "(isNil (Abs_stream [0 \<mapsto> 2::int, 2 \<mapsto> 4, 4 \<mapsto> 7])) \<lbrakk>3\<rbrakk> = (Some True)" 
       by(simp add: isNil_def sel_def Abs_stream_inverse' null_option_def)
value  "(Abs_stream(\<lambda>x . Some (int x)))\<lbrakk>5\<rbrakk>"


end