theory Lustre_types
imports Main
begin



subsection{* Infrastructure on Null-Types (bu's way) *}

class      null = 
   fixes   null :: "'a"
   assumes nonEmpty : "\<exists> x. x \<noteq> null"

abbreviation is_def :: "'a::null \<Rightarrow> bool"
where       "is_def X \<equiv> (X \<noteq> null)"

instantiation   option  :: (type)null
begin
   definition null_option_def[simp]: "(null::'a option) \<equiv> (None::'a option)"
   instance proof show "\<exists>(x::'a option). x \<noteq> null"
               by(rule_tac x="Some x" in exI, simp)
            qed
end

(* Some types to play with : *)

type_synonym Integer  = "int option"
type_synonym Bool     = "bool option"


instantiation   prod  :: (null,null)null
begin
   definition null_prod_def[simp]: "(null::'a \<times> 'b) \<equiv> (null, null)"
   instance proof show "\<exists>(x::'a \<times> 'b). x \<noteq> null"
                  apply(insert nonEmpty)
                  apply(erule_tac P="\<lambda>x::'a. x \<noteq> null" in exE)
                  apply(rule_tac x="(x,null)" in exI)
                  apply(simp add: null_prod_def)
                  done
            qed
end



text{* Streams are represented by functions over natural numbers to value domains
       that must contain a distinguished null element, in order to model the 
       explicit undefinedness and strictness principles of the Lustre language. 

       For Isabelle-technical reasons, the natural definition
       \begin{inlineisar}
          typedef      'a stream = "UNIV::(nat \<Rightarrow> 'a::null) set" by (auto)
       \end{inlineisar} 
       is slightly liberalized to:
 *}


typedef 'a stream = "UNIV::(nat \<Rightarrow> 'a) set" by (auto)

text{* This too liberal typedefinition is implicitly restricted by all operations 
       on streams that assume it to be a stream possibly containing null values.*}

lemmas Abs_stream_inverse'= Abs_stream_inverse[simplified]

setup_lifting type_definition_stream


lift_definition isNil :: "'a::null stream \<Rightarrow> Bool stream" is
"\<lambda> S n. Some (S n = null)" .

lift_definition nil  :: "'a::null stream" is
"\<lambda> n. null" .


text{* The nil-stream is also the null element in the space of streams *}
(*TODO:  identify nil and null ? *)
instantiation stream :: (null) null
begin

definition null_stream :: "'a stream" where "null = nil"
instance 
  proof 
    show "\<exists>x::('a::null) stream. x \<noteq> null"
    apply(rule_tac x="Abs_stream(\<lambda>n. if n=0 then null else @x. is_def x)" in exI)
    unfolding null_stream_def nil_def
    apply(subst Abs_stream_inject, auto)
    apply(drule_tac x="Suc 0" in fun_cong, insert nonEmpty, auto)
    apply(erule contrapos_pp)
    apply(rule Hilbert_Choice.someI)
    by assumption
  qed
end





end