theory counter
imports lustre
begin

text{* 
We try to encode the following standard example from the Lustre - Tutorial:
\begin{verbatim}
node COUNTER (init, incr : int; X, reset : bool) returns (C : int);
var PC : int;
let 
    PC = init -> pre C;
    C  = if reset then init
         else if X then (PC + incr)
              else PC;
tel

\end{verbatim}
*}

type_synonym COUNTER_inputs = "Integer \<times> Integer \<times> Bool \<times> Bool"

type_synonym COUNTER_outputs = "Integer"

type_synonym COUNTER_state = "Integer"


fun    COUNTER_block::"(COUNTER_inputs stream \<times> COUNTER_state stream \<times> COUNTER_outputs stream) \<Rightarrow> 
                                               (COUNTER_state stream \<times> COUNTER_outputs stream)" 
where "COUNTER_block (inp, st, out) =
(let PC = st;
     \<lbrakk>init, incr, X, reset\<rbrakk> = inp;
     C = out;     
     PC' = init -> pre (C);                       (* construction of state variables *) 
     st' = PC';                                   (* constructions of internal state-stream *)
     C'  = if reset then init                     (* construction of output variables *)
           else if X then PC' + incr
                else PC'
                endif 
           endif;
     out' = C'                                    (* construction of output stream*)
 in (st',out'))"



definition  COUNTER :: "Integer stream \<times> Integer stream \<times> Bool stream \<times> Bool stream \<Rightarrow> 
                        Integer stream" 
where      "COUNTER \<equiv> (\<lambda>(init, incr, X, reset). (loop COUNTER_block \<lbrakk>init, incr, X, reset\<rbrakk>))"

(* Example for the construction of an input stream ... *)
value "mk_COUNTER_inputs_stream (\<lbrace>[10,12,14,16]\<rbrace>,
                                 \<lbrace>[4,5,6,7]\<rbrace>,
                                 \<lbrace>[True,False,True,True]\<rbrace>,
                                 \<lbrace>[True,False,True,True]\<rbrace>)"

(* TODO : Configure Code Generator *)

end
