theory proofs
imports verif_speed
begin

lemma strict_unary_dest: 
"strict_unary a b = Some c \<Longrightarrow> \<exists> x. b = Some x \<and> a x = c"
by (case_tac b, auto)

lemma strict_binary_dest: 
"strict_binary a b c = Some d \<Longrightarrow> \<exists> x y. b = Some x \<and> c = Some y \<and> a x y = d"
apply (case_tac b, auto)
by (drule strict_unary_dest)

lemma strict_unary_none_dest: 
"strict_unary a b = None \<Longrightarrow> b = None"
by (case_tac b, auto)

lemma strict_binary_none_dest: 
"strict_binary a b c = None \<Longrightarrow> b = None \<or> c = None"
apply (case_tac b, auto)
by (drule strict_unary_none_dest)

lemmas strict_dest = strict_unary_dest strict_binary_dest strict_unary_none_dest strict_binary_none_dest

lemma strict_unary_intro: 
"\<And>x. b = Some x \<and> a x = c \<Longrightarrow> strict_unary a b = Some c"
by (case_tac b, auto)

lemma strict_binary_intro: 
"\<And>x y. b = Some x \<and> c = Some y \<and> a x y = d \<Longrightarrow> strict_binary a b c = Some d"
by (case_tac b, auto)

lemma strict_unary_none_intro: 
"b = None \<Longrightarrow> strict_unary a b = None"
by (case_tac b, auto)

lemma strict_binary_none_intro: 
"b = None \<or> c = None \<Longrightarrow> strict_binary a b c = None"
by (case_tac b, auto)

lemmas strict_intro = strict_unary_intro strict_binary_intro strict_unary_none_intro strict_binary_none_intro

lemma option_intro:
  assumes A: "x=None \<Longrightarrow> F P"
  and     B: "\<And>y. x=Some y \<Longrightarrow> F (Q y)"
  shows "F (case x of None \<Rightarrow> P | Some y \<Rightarrow> Q y)"
by (case_tac x, auto simp: A B)


lemma option_elim:
  assumes A: "F (case x of None \<Rightarrow> P | Some y \<Rightarrow> Q y)"
  and     B: "x=None \<Longrightarrow> F P \<Longrightarrow> R None"
  and     C: "\<And>y. x=Some y \<Longrightarrow> F (Q y) \<Longrightarrow> R (Some y)"
  shows "R x"
apply (insert A)
apply (case_tac x, auto simp: B C)
done



lemma verif_speed : "\<forall> inp. assert (verif_speed inp)"
apply (auto simp: assert_def verif_speed_def)
apply (rule loop_induct)
apply (auto split: option.split option.split_asm Product_Type.split_split Product_Type.split_split_asm)

apply (clarsimp simp add: speed_def Let_def)+
(* Here we stop; this is no longer provable but actually this is not necessarily a counter-example *)

oops


lemma counter_tmp:
"COUNTER (0, x, y, false)\<lbrakk>Suc k\<rbrakk> = None \<longrightarrow> (x\<lbrakk>Suc k\<rbrakk>=None \<or> y\<lbrakk>Suc k\<rbrakk>=None \<or> COUNTER (0, x, y, false)\<lbrakk>k\<rbrakk> = None)"
apply (induct k)
apply (auto simp: COUNTER_def Let_def loop_def
            split: option.split option.split_asm
                   Product_Type.split_split Product_Type.split_split_asm)
done

lemma verif_speed2 : "\<forall> inp. assert (verif_speed2 inp)"
proof (auto simp: assert_def,auto split: option.split)
      fix a b n
      show "\<exists>y. verif_speed2 (a, b)\<lbrakk>n\<rbrakk> = Some y" (* proof of pointwise definedness *)
           unfolding verif_speed2_def
           proof (induct n)
           case 0 show ?case
            by (auto simp: Let_def speed_def split: option.split)
           case (Suc k) from Suc.hyps show ?case
                 apply (auto split: option.split option.split_asm
                                    Product_Type.split_split Product_Type.split_split_asm
                             simp: Let_def loop_simp)
                 apply (auto split: option.split option.split_asm
                                    Product_Type.split_split Product_Type.split_split_asm
                             simp: speed_def Let_def)
                 apply (case_tac "loop\<^sub>0 speed_block \<lbrakk>a, b\<rbrakk> k")
                 apply (auto split: option.split option.split_asm
                                    Product_Type.split_split Product_Type.split_split_asm
                             simp: speed_def Let_def)
                 apply (case_tac k, simp_all)
                 apply (clarsimp split: option.split option.split_asm 
                                   Product_Type.split_split Product_Type.split_split_asm 
                            simp: Let_def verif_speed2_def speed_def COUNTER_def)
                 apply (clarsimp split: option.split option.split_asm 
                                   Product_Type.split_split Product_Type.split_split_asm 
                            simp: Let_def verif_speed2_def speed_def COUNTER_def)
                 apply (case_tac x2, simp_all)
                 apply (clarsimp split: option.split option.split_asm HOL.split_if HOL.split_if_asm
                                   Product_Type.split_split Product_Type.split_split_asm 
                            simp: Let_def)
                 apply (drule counter_tmp[rule_format])
                 apply (clarsimp split: option.split option.split_asm HOL.split_if HOL.split_if_asm
                                   Product_Type.split_split Product_Type.split_split_asm 
                            simp: Let_def COUNTER_def)
                 apply (clarsimp split: option.split option.split_asm HOL.split_if HOL.split_if_asm
                                   Product_Type.split_split Product_Type.split_split_asm 
                            simp: Let_def COUNTER_def speed_def)
                 apply (case_tac nat, simp_all)
                 apply (clarsimp split: option.split option.split_asm HOL.split_if HOL.split_if_asm
                                   Product_Type.split_split Product_Type.split_split_asm 
                            simp: Let_def COUNTER_def)
                 apply (clar simp split: option.split option.split_asm HOL.split_if HOL.split_if_asm
                                   Product_Type.split_split Product_Type.split_split_asm 
                            simp: Let_def COUNTER_def)
                 apply (clarsimp simp: Let_def COUNTER_def)
                 apply (split Product_Type.split_split_asm)+
                 apply (clarsimp)
                 apply (split option.split_asm)+
                 apply (clarsimp)+
                 apply (split option.split_asm)+
                 apply (clarsimp)+
                 apply (split option.split_asm)+
                 apply (clarsimp)+
                 apply (split option.split_asm)+
                 apply (clarsimp)+
                 apply (case_tac x2j, simp_all)
                 apply (split option.split_asm)+
                 apply (clarsimp simp: Let_def)+
                 apply (case_tac "x2o \<or> x2q", simp_all)
                 apply (split option.split_asm)+
                 apply (clarsimp)+
                 apply (split option.split_asm)+
                 apply (clarsimp)+
                 apply (case_tac "x2o \<and> \<not> x2q", simp_all)
                 apply (split option.split_asm)+
                 apply (clarsimp)+

oops
                 apply(clarsi split: option.split option.split_asm 
                                   Product_Type.split_split Product_Type.split_split_asm 
                            simp: Let_def verif_speed2_def speed_def COUNTER_def loop_def
                                  speed_inputs.splits verif_speed_inputs.splits
                                  speed_inputs.defs verif_speed_inputs.defs)
                 apply(case_tac k, simp_all split: option.split_asm Product_Type.split_split_asm)
                 apply(au to split: option.split option.split_asm 
                                   Product_Type.split_split Product_Type.split_split_asm 
                            simp: Let_def verif_speed2_def speed_def COUNTER_def loop_def)
                 apply(case_tac k, simp_all split: option.split_asm Product_Type.split_split_asm)
                 apply(au to split: option.split option.split_asm 
                                   Product_Type.split_split Product_Type.split_split_asm 
                            simp: Let_def verif_speed2_def speed_def COUNTER_def loop_def)
                 apply(case_tac nat, simp_all split: option.split_asm Product_Type.split_split_asm)
                 apply(auto split: option.split option.split_asm 
                                   Product_Type.split_split Product_Type.split_split_asm 
                            simp: Let_def verif_speed2_def speed_def COUNTER_def loop_def)
                 apply(case_tac k, simp_all split: option.split_asm Product_Type.split_split_asm)
                 apply(auto split: option.split option.split_asm 
                                   Product_Type.split_split Product_Type.split_split_asm 
                            simp: Let_def verif_speed2_def speed_def COUNTER_def loop_def)
                 apply(case_tac nat, simp_all split: option.split_asm Product_Type.split_split_asm)
                 apply(auto split: option.split option.split_asm 
                                   Product_Type.split_split Product_Type.split_split_asm 
                            simp: Let_def verif_speed2_def speed_def COUNTER_def loop_def)
                 apply(case_tac k, simp_all split: option.split_asm Product_Type.split_split_asm)
                 apply(auto split: option.split option.split_asm 
                                   Product_Type.split_split Product_Type.split_split_asm 
                            simp: Let_def verif_speed2_def speed_def COUNTER_def loop_def)
                 apply(case_tac k, simp_all split: option.split_asm Product_Type.split_split_asm)
                 apply(auto split: option.split option.split_asm 
                                   Product_Type.split_split Product_Type.split_split_asm 
                            simp: Let_def verif_speed2_def speed_def COUNTER_def loop_def)
                 apply(case_tac nat, simp_all split: option.split_asm Product_Type.split_split_asm)
                 apply(auto split: option.split option.split_asm 
                                   Product_Type.split_split Product_Type.split_split_asm 
                            simp: Let_def verif_speed2_def speed_def COUNTER_def loop_def)
                 apply(case_tac nata, simp_all split: option.split_asm Product_Type.split_split_asm)
                 apply(auto split: option.split option.split_asm 
                                   Product_Type.split_split Product_Type.split_split_asm 
                            simp: Let_def verif_speed2_def speed_def COUNTER_def loop_def)

                 apply(auto split: option.split option.split_asm 
                                   Product_Type.split_split Product_Type.split_split_asm 
                            simp: Let_def)
                 unfolding  COUNTER_def loop_def
                 apply(auto split: option.split option.split_asm 
                                   Product_Type.split_split Product_Type.split_split_asm 
                            simp: Let_def) 
               sorry
     qed
next
      fix a b n x2
      show "Rep_stream (verif_speed2 (a, b)) n = Some x2 \<Longrightarrow> x2"
           proof (induct n arbitrary: a b)
              case 0  show ?case
                 unfolding  verif_speed2_def loop_def
                 apply(auto split: option.split Product_Type.split_split simp: Let_def)
                 sorry
              case (Suc k) from Suc.hyps show ?case
                 unfolding  verif_speed2_def loop_def
                 apply(auto split: option.split Product_Type.split_split simp: Let_def)
                 unfolding  speed_def loop_def
                 apply(auto split: option.split Product_Type.split_split simp: Let_def) 
                 apply(case_tac k, simp_all split: option.split_asm Product_Type.split_split_asm)
                 unfolding  COUNTER_def loop_def
                 apply(auto split: option.split Product_Type.split_split simp: Let_def) 
               sorry

      sorry
qed


apply (rule strict_intro, auto)
apply (clarsimp simp: verif_speed2_def extract_stream_def)
apply (rule loop_induct, auto)
apply (rule option_intro, auto)
apply (clarsimp simp: verif_speed2_block_def Let_def make_stream_def speed_def loop_def)
apply (clarsimp simp: verif_speed2_block_def loop_def Let_def make_stream_def speed_def extract_stream_def)

apply (rule option_intro, auto)
apply (clarsimp simp: verif_speed2_block_def verif_speed_outputs.defs verif_speed_inputs.defs
                  Abs_stream_inverse[simplified] Let_def empty_verif_speed_state_def make_stream_def
                  sel_def verif_speed_state.defs extract_stream_def)
apply (subst (asm) Rep_stream_inverse)+
apply (auto simp: speed_def speed_inputs.defs empty_verif_speed_outputs_def verif_speed_outputs.defs
                  Abs_stream_inverse[simplified] verif_speed_state.defs Let_def extract_stream_def)

apply (auto simp: verif_speed2_block_def Let_def make_stream_def null_def sel_def
                  empty_verif_speed_state_def empty_verif_speed_outputs_def extract_stream_def
                  Abs_stream_inverse[simplified] not_def speed_def)
apply (rule strict_intro, auto)
apply (rule strict_intro, auto)
apply (rule option_intro, auto)

apply (case_tac n, simp_all)
apply (clarsimp simp: loop_def fun_eq_iff)
apply (case_tac a, clarsimp)
apply (case_tac b, clarsimp)
apply (clarsimp simp: verif_speed2_block_def Let_def make_stream_def null_def sel_def fby_def
                  empty_verif_speed_state_def empty_verif_speed_outputs_def extract_stream_def
                  true_def Abs_stream_inverse[simplified] not_def speed_def speed_block_def
                  COUNTER_def)
apply (case_tac nat, simp_all)
apply (clarsimp simp: loop_def fun_eq_iff)
apply (clarsimp simp: verif_speed2_block_def Let_def make_stream_def null_def sel_def fby_def
                  empty_verif_speed_state_def empty_verif_speed_outputs_def extract_stream_def
                  true_def Abs_stream_inverse[simplified] not_def speed_def speed_block_def)
apply (clarsimp simp: loop_def fun_eq_iff)
apply (clarsimp simp: COUNTER_def Let_def make_stream_def null_def sel_def fby_def If_def
                  empty_verif_speed_state_def empty_verif_speed_outputs_def extract_stream_def
                  true_def Abs_stream_inverse[simplified] not_def speed_def speed_block_def)
apply (case_tac n, simp_all)
apply (clarsimp simp: loop_def fun_eq_iff)
apply (clarsimp simp: verif_speed2_block_def Let_def make_stream_def null_def sel_def fby_def
                  empty_verif_speed_state_def empty_verif_speed_outputs_def extract_stream_def
                  true_def Abs_stream_inverse[simplified] not_def speed_def speed_block_def)



oops


lemma temp2: 
"Rep_stream (true -> ((not (late beacon second)) and pre (early beacon second))) (Suc na) = Some True \<longrightarrow> P na"
apply (rule nat_induct)
apply (clarsimp simp: true_def fby_def Abs_stream_inverse[simplified] and_def)
apply (case_tac "(Rep_stream (not late beacon second)) (Suc 0)", auto)
apply (case_tac "(Rep_stream (pre (early beacon second))) (Suc 0)", auto)
apply (clarsimp simp: Abs_stream_inverse[simplified] not_def pre_def)
apply (case_tac "(Rep_stream (late beacon second)) (Suc 0)", auto)
apply (clarsimp simp: Abs_stream_inverse[simplified] late_def early_def)

apply (clarsimp simp: true_def fby_def Abs_stream_inverse[simplified] and_def)
apply (case_tac "(Rep_stream (not late beacon second)) (Suc n)", auto)
apply (case_tac "(Rep_stream (not late beacon second)) (Suc (Suc n))", auto)
apply (case_tac "(Rep_stream (pre (early beacon second))) (Suc (Suc n))", auto)
apply (clarsimp simp: Abs_stream_inverse[simplified] not_def pre_def)
apply (case_tac "(Rep_stream (late beacon second)) (Suc n)", auto)
apply (case_tac "(Rep_stream (late beacon second)) (Suc (Suc n))", auto)
apply (clarsimp simp: Abs_stream_inverse[simplified] late_def early_def)

apply (case_tac "(Rep_stream (not late beacon second)) (Suc (Suc n))", auto)
apply (case_tac "(Rep_stream (pre (early beacon second))) (Suc (Suc n))", auto)
apply (clarsimp simp: Abs_stream_inverse[simplified] not_def pre_def)
apply (case_tac "(Rep_stream (late beacon second)) (Suc n)", auto)
apply (case_tac "(Rep_stream (late beacon second)) (Suc (Suc n))", auto)
apply (case_tac "(Rep_stream (early beacon second)) n", auto)
apply (clarsimp simp: Abs_stream_inverse[simplified] late_def early_def)
apply (clarsimp simp: Abs_stream_inverse[simplified] late_def early_def)
apply (case_tac "early_f  diff beacon second > 0  diff beacon second >= 10 n")
apply (clarsimp simp: Abs_stream_inverse[simplified] sel_def)
apply (case_tac "(Rep_stream ((diff beacon second) > 0)) (Suc n)")
apply (clarsimp simp: Abs_stream_inverse[simplified] diff_def COUNTER_def gt_def)
apply (case_tac "late_f ((diff beacon second) < 0) ((diff beacon second) <= - 10) n")
apply (clarsimp simp: Abs_stream_inverse[simplified] COUNTER_def gt_def)

apply (clarsimp simp: Abs_stream_inverse[simplified] sel_def diff_def COUNTER_def false_def id_def
                      lt_def gt_def le_def ge_def)

apply (insert Rep_Abs_stream)

thm Abs_stream_inverse[simplified]

apply (case_tac "(Rep_stream (not late beacon second)) (Suc na)", auto)
apply (case_tac "(Rep_stream (pre (early beacon second))) (Suc na)", auto)
apply (auto simp: Abs_stream_inverse[simplified] not_def pre_def)
apply (case_tac "(Rep_stream (late beacon second)) (Suc na)", auto)
apply (auto simp: Abs_stream_inverse[simplified] late_def)
apply (case_tac " late_f ((diff beacon second) < 0)  ((diff beacon second) <= - 10) na")
apply (auto simp: Abs_stream_inverse[simplified] early_def)



apply (auto simp: true_def fby_def Abs_stream_inverse[simplified] not_def late_def pre_def and_def 
          early_def sel_def diff_def COUNTER_def or_def id_def incr_def ge_def plus_stream_def
          zero_stream_def false_def lt_def uminus_stream_def one_stream_def)
oops





lemma if_dest:
"Rep_stream (if a then b else c) n = Some d \<Longrightarrow> 
  \<exists> x. ((Rep_stream a) n) = Some x \<and> (if x then (Rep_stream b) n else (Rep_stream c) n) = Some d"
apply (clarsimp simp: If_def Abs_stream_inverse[simplified])
apply (case_tac "Rep_stream a n", simp_all)
apply (rule_tac x=aa in exI, auto)
done

lemma if_none_dest:
"Rep_stream (if a then b else c) n = None \<Longrightarrow> 
  ((Rep_stream a) n) \<noteq> None \<longrightarrow> 
    (if the ((Rep_stream a) n) then (Rep_stream b) n else (Rep_stream c) n) = None"
apply (clarsimp simp: If_def Abs_stream_inverse[simplified])
apply (case_tac "Rep_stream a n", simp_all)
apply (case_tac "aa", simp_all)
done

lemmas If_dest = if_dest if_none_dest




lemma "Rep_stream  diff beacon second <= - 10 (Suc n) \<noteq> None"
apply (case_tac "Rep_stream  diff beacon second <= - 10 (Suc n)")
apply (auto simp: le_def Abs_stream_inverse[simplified])
apply (drule strict_dest)+
apply (auto simp: uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: diff_def COUNTER_def Abs_stream_inverse[simplified])
done

lemma "Rep_stream (late beacon second) (Suc n) = Some False \<Longrightarrow>
       Rep_stream (late beacon second) (Suc (Suc n)) = Some True \<Longrightarrow> False"
apply (auto simp: late_def sel_def Abs_stream_inverse[simplified])
apply (case_tac "Rep_stream  diff beacon second <= - 10 (Suc (Suc n))")
apply (auto simp: le_def Abs_stream_inverse[simplified])
apply (drule strict_dest)+
apply (auto simp: uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: diff_def COUNTER_def Abs_stream_inverse[simplified])
apply (drule strict_dest)+
apply (auto simp: uminus_stream_def Abs_stream_inverse[simplified])
apply (case_tac "late_f  diff beacon second < 0 (Abs_stream (strict_binary op \<le> (Rep_stream (diff beacon second)) (Rep_stream (- 10)))) n")
apply (auto simp: sel_def Abs_stream_inverse[simplified])
apply (case_tac "Rep_stream  diff beacon second < 0 (Suc n)")
apply (auto simp: lt_def Abs_stream_inverse[simplified])
apply (drule strict_dest)+
apply (clarsimp simp: COUNTER_def diff_def Abs_stream_inverse[simplified])
apply (drule strict_dest)+
apply (clarsimp simp: diff_def COUNTER_def Abs_stream_inverse[simplified])

oops



lemma nat_2_induct:
  assumes A:"P 0" and B:"P (Suc 0)" and C:"\<And>n. P n \<Longrightarrow> P (Suc n) \<Longrightarrow> P (Suc (Suc n))"
  shows "P n" 
apply (rule nat_induct, simp add: A)
apply (subgoal_tac "P n & P (Suc n)")
apply (simp)
apply (rule_tac P="\<lambda>n. P n & P (Suc n)" in nat_induct)
apply (auto simp: A B C)
done


lemma 
"n>0 &
      (late_f ((diff beacon second) < 0) ((diff beacon second) <= - 10) n) \<longrightarrow>
       ~(early_f  ((diff beacon second) > 0) ((diff beacon second) >= 10) (n - Suc 0))"
apply (rule nat_2_induct)
apply (induct_tac n, simp)
apply (auto)
apply (subst (asm) diff_def) back back
apply (subst (asm) le_def) back
apply (auto simp: sel_def uminus_stream_def COUNTER_def id_def Abs_stream_inverse[simplified])
apply (case_tac "Rep_stream beacon or second (Suc n)") defer
apply (clarsimp simp: or_def Abs_stream_inverse[simplified])
apply (drule strict_dest)+
apply (clarsimp simp: or_def Abs_stream_inverse[simplified])
apply (case_tac "x \<or> y", simp_all)
apply (case_tac "(beacon or second)\<lbrakk>n\<rbrakk>", simp_all)

oops


lemma temp_lemma1:
"Rep_stream  ((diff beacon second) <= - 10) (Suc (Suc 0)) = Some True \<Longrightarrow>
 Rep_stream  ((diff beacon second) >= 10) (Suc 0) = Some True \<Longrightarrow> 
 Rep_stream  ((diff beacon second) <= - 10) (Suc 0) = Some ab \<Longrightarrow> ab"
apply (auto simp: le_def ge_def Abs_stream_inverse[simplified])
apply (drule strict_dest)+
apply (auto simp: uminus_stream_def diff_def COUNTER_def  Abs_stream_inverse[simplified])
apply (case_tac "beacon or second\<lbrakk>Suc (Suc 0)\<rbrakk>")
apply (auto simp: Abs_stream_inverse[simplified])
apply (case_tac "beacon or second\<lbrakk>Suc 0\<rbrakk>")
apply (auto simp: Abs_stream_inverse[simplified])
apply (case_tac "beacon or second\<lbrakk>0\<rbrakk>")
apply (auto simp: Abs_stream_inverse[simplified])
apply (clarsimp simp: sel_def or_def plus_stream_def Abs_stream_inverse[simplified])
apply (erule option_caseE)+
apply (auto)
apply (drule strict_dest)+
apply (subst (asm) option.split_asm)
apply (clarsimp)
apply (subst (asm) option.split_asm)
apply (clarsimp)
apply (erule disjE) back back back
apply (erule disjE) back back back
apply (clarsimp)
apply (drule strict_dest)+
apply (clarsimp)
apply (clarsimp)
apply (drule strict_dest)+
apply (clarsimp)
apply (erule disjE) back back back
apply (clarsimp)
apply (drule strict_dest)+
apply (clarsimp)
apply (clarsimp)
apply (drule strict_dest)+
apply (clarsimp)
apply (subst (asm) option.split_asm)
apply (clarsimp)
apply (subst (asm) option.split_asm)
apply (clarsimp)
apply (erule disjE) back back back
apply (erule disjE) back back back
apply (clarsimp)
apply (drule strict_dest)+
apply (clarsimp)
apply (clarsimp)
apply (drule strict_dest)+
apply (clarsimp)
apply (erule disjE) back back back
apply (clarsimp)
apply (drule strict_dest)+
apply (clarsimp)
apply (clarsimp)
apply (drule strict_dest)+
apply (clarsimp)
apply (subst (asm) option.split_asm)
apply (clarsimp)
apply (subst (asm) option.split_asm)
apply (clarsimp)
apply (erule disjE) back back
apply (erule disjE) back back
apply (clarsimp)
apply (drule strict_dest)+
apply (clarsimp)
apply (clarsimp)
apply (drule strict_dest)+
apply (clarsimp)
apply (erule disjE) back back
apply (clarsimp)
apply (drule strict_dest)+
apply (clarsimp)
apply (clarsimp)
apply (drule strict_dest)+
apply (clarsimp)
apply (subst (asm) option.split_asm)
apply (clarsimp)
apply (subst (asm) option.split_asm)
apply (clarsimp)
apply (erule disjE) back
apply (erule disjE) back
apply (clarsimp)
apply (drule strict_dest)+
apply (clarsimp)
apply (clarsimp)
apply (drule strict_dest)+
apply (clarsimp)
apply (erule disjE) back
apply (clarsimp)
apply (drule strict_dest)+
apply (clarsimp)
apply (clarsimp)
apply (clarsimp simp: sel_def or_def plus_stream_def Abs_stream_inverse[simplified])
apply (subst (asm) option.split_asm)
apply (clarsimp)
apply (subst (asm) option.split_asm)
apply (clarsimp)
apply (erule disjE)
apply (erule disjE)
apply (clarsimp)
apply (drule strict_dest)+
apply (clarsimp)
apply (clarsimp)
apply (drule strict_dest)+
apply (clarsimp)
apply (erule disjE)
apply (clarsimp)
apply (drule strict_dest)+
apply (clarsimp simp: incr_def)
apply (drule If_dest)
apply (clarsimp simp: and_def not_def Abs_stream_inverse[simplified])
apply (drule strict_dest)+
apply (clarsimp simp: and_def not_def Abs_stream_inverse[simplified])
apply (drule strict_dest)+
apply (case_tac "xa \<and> ya", simp_all)
apply (drule If_dest)
apply (clarsimp simp: and_def not_def Abs_stream_inverse[simplified])
apply (case_tac "beacon or second\<lbrakk>0\<rbrakk>")
apply (auto simp: Abs_stream_inverse[simplified])
apply (clarsimp simp: sel_def or_def plus_stream_def Abs_stream_inverse[simplified])
apply (subst (asm) option.split_asm)
apply (clarsimp)
apply (subst (asm) option.split_asm)
apply (clarsimp)
apply (erule disjE)
apply (erule disjE)
apply (clarsimp)
apply (drule strict_dest)+
apply (clarsimp)
apply (clarsimp)
apply (drule strict_dest)+
apply (clarsimp)
apply (erule disjE)
apply (clarsimp)
apply (drule strict_dest)+
apply (clarsimp simp: incr_def)
apply (drule If_dest)
apply (clarsimp simp: and_def not_def Abs_stream_inverse[simplified])
apply (drule strict_dest)+
apply (clarsimp simp: and_def not_def Abs_stream_inverse[simplified])
apply (drule strict_dest)+
apply (case_tac "xa \<and> ya", simp_all)
apply (drule If_dest)
apply (clarsimp simp: and_def not_def Abs_stream_inverse[simplified])
apply (case_tac "x \<and> ~xa", simp_all)
apply (clarsimp simp: and_def not_def uminus_stream_def Abs_stream_inverse[simplified])
apply (case_tac "beacon or second\<lbrakk>Suc 0\<rbrakk>")
apply (auto simp: Abs_stream_inverse[simplified])
done



lemma verif_speed2 : "\<forall> beacon second. assert ((ok' beacon second) =s true)"
apply (auto simp: assert_def eq_def ok'_def)
apply (rule nat_2_induct)
apply (clarsimp simp: not_def and_def pre_def true_def early_def late_def
                  fby_def sel_def Abs_stream_inverse[simplified])
apply (clarsimp simp: not_def and_def pre_def true_def early_def late_def
                  fby_def sel_def Abs_stream_inverse[simplified])
apply (clarsimp simp: not_def and_def pre_def true_def early_def late_def
                  fby_def sel_def Abs_stream_inverse[simplified])
apply (case_tac na)
apply (clarsimp simp: not_def and_def pre_def true_def early_def late_def 
                   fby_def sel_def Abs_stream_inverse[simplified])
apply (rule)+
apply (erule option_caseE)+
apply (clarsimp simp: le_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: le_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: le_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (erule option_caseE)+
apply (clarsimp simp: lt_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: lt_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (erule option_caseE)+
apply (clarsimp simp: ge_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: lt_def ge_def le_def diff_def COUNTER_def uminus_stream_def 
                      Abs_stream_inverse[simplified])
apply (rule)+
apply (erule option_caseE)+
apply (clarsimp simp: le_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: le_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (erule notE)
apply (erule option_caseE)+
apply (clarsimp simp: ge_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (case_tac "Rep_stream  ((diff beacon second) <= - 10) (Suc 0)")
apply (clarsimp simp: le_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: COUNTER_def Abs_stream_inverse[simplified])
apply (drule temp_lemma1)
apply (simp_all)

apply (clarsimp simp: not_def and_def pre_def true_def early_def late_def 
                   fby_def sel_def Abs_stream_inverse[simplified])
apply (rule)+
apply (erule option_caseE)+
apply (clarsimp simp: lt_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: lt_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: lt_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: lt_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: gt_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (erule option_caseE)+
apply (clarsimp simp: lt_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: lt_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: lt_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: lt_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (erule option_caseE)+
apply (clarsimp simp: lt_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: lt_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: lt_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (erule option_caseE)+
apply (clarsimp simp: lt_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: lt_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (erule option_caseE)+
apply (clarsimp simp: ge_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: lt_def ge_def le_def diff_def COUNTER_def uminus_stream_def 
                      Abs_stream_inverse[simplified])
apply (rule)+
apply (erule option_caseE)+
apply (clarsimp simp: le_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: le_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: lt_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: gt_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (erule option_caseE)+
apply (clarsimp simp: lt_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: lt_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: lt_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (erule option_caseE)+
apply (clarsimp simp: le_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: le_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (erule option_caseE)+
apply (clarsimp simp: gt_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (erule notE)
apply (case_tac "Rep_stream  ((diff beacon second) <= - 10) (Suc 0)")
apply (clarsimp simp: le_def diff_def COUNTER_def uminus_stream_def Abs_stream_inverse[simplified])
apply (clarsimp simp: COUNTER_def Abs_stream_inverse[simplified])
apply (rule temp_lemma1)
apply (simp_all)




oops

end