theory Lustre_consistency
imports Lustre_basics
begin

subsection{* Computational Consistency. *}

text{*  Lustre stream operations can depend on the history in a stream; in principle, the output
of a computationally consistent computation can depend on all computations in this stream before. 
A simple way to characterize this is by a form of \emph{generalized equality} or:
equality of two streams up to some point @{term n} in the future.
*}

definition equal_upto :: "'a::null stream \<Rightarrow> nat \<Rightarrow> 'a stream \<Rightarrow> bool" ("_ \<triangleq>\<^sub>\<lbrakk>\<^bsub>_\<^esub>\<^sub>\<rbrakk> _" [40,0,40] 60)
where "S \<triangleq>\<^sub>\<lbrakk>\<^bsub>n\<^esub>\<^sub>\<rbrakk> T \<equiv> (\<forall> k \<le> n. (S\<lbrakk>k\<rbrakk>) = (T\<lbrakk>k\<rbrakk>))" 

lemma equal_upto_refl : "S \<triangleq>\<^sub>\<lbrakk>\<^bsub>n\<^esub>\<^sub>\<rbrakk> S"
by(simp add:equal_upto_def)

lemma equal_upto_sym : "S \<triangleq>\<^sub>\<lbrakk>\<^bsub>n\<^esub>\<^sub>\<rbrakk> T \<Longrightarrow> T \<triangleq>\<^sub>\<lbrakk>\<^bsub>n\<^esub>\<^sub>\<rbrakk> S"
by(simp add:equal_upto_def)

lemma equal_upto_trans : "S \<triangleq>\<^sub>\<lbrakk>\<^bsub>n\<^esub>\<^sub>\<rbrakk> T \<Longrightarrow> T \<triangleq>\<^sub>\<lbrakk>\<^bsub>n\<^esub>\<^sub>\<rbrakk> U \<Longrightarrow> S \<triangleq>\<^sub>\<lbrakk>\<^bsub>n\<^esub>\<^sub>\<rbrakk> U"
by(simp add:equal_upto_def)

lemma equal_upto_mono : "k \<le> n \<Longrightarrow> S \<triangleq>\<^sub>\<lbrakk>\<^bsub>n\<^esub>\<^sub>\<rbrakk> T \<Longrightarrow> S \<triangleq>\<^sub>\<lbrakk>\<^bsub>k\<^esub>\<^sub>\<rbrakk> T"
by(simp add:equal_upto_def)



text{* Now, if a term context @{term P} has the property that two streams which are equal upto
 @{term n} satisfy the Leipnitz property upto @{term n}, i.e. they enjoy substitutivity. *}

definition cc :: " ('a::null stream \<Rightarrow> 'b::null stream) \<Rightarrow> bool" ("\<langle>_\<rangle>")
where     "\<langle>P\<rangle> \<equiv> \<forall> n S T.  S \<triangleq>\<^sub>\<lbrakk>\<^bsub>n\<^esub>\<^sub>\<rbrakk> T \<longrightarrow> P(S) \<triangleq>\<^sub>\<lbrakk>\<^bsub>n\<^esub>\<^sub>\<rbrakk> P(T)" 


text{* As said above, Lustre stream operations can depend on the history in a stream.
However, most Lustre library operators do not depend on the history in a stream. This fact 
can be exploited in proofs. We therefore introduce the concept of \emph{strong computationally 
consistent} stream functions, which are exactly those that can be constructed pointwise. *}

definition scc :: " ('a::null stream \<Rightarrow> 'b::null stream) \<Rightarrow>  bool"  ("!\<langle>_\<rangle>")
where     "!\<langle>P\<rangle> \<equiv> \<forall> n S T. (S\<lbrakk>n\<rbrakk>) = (T\<lbrakk>n\<rbrakk>) \<longrightarrow> ((P S)\<lbrakk>n\<rbrakk>) = ((P T)\<lbrakk>n\<rbrakk>)" 


   
lemma equal_upto_substI : "\<langle>P\<rangle> \<Longrightarrow> S \<triangleq>\<^sub>\<lbrakk>\<^bsub>n\<^esub>\<^sub>\<rbrakk> T \<Longrightarrow> P(S) \<triangleq>\<^sub>\<lbrakk>\<^bsub>n\<^esub>\<^sub>\<rbrakk> P(T)"
using cc_def by auto

lemma sccD : "!\<langle>P\<rangle> \<Longrightarrow> (S\<lbrakk>n\<rbrakk>) = (T\<lbrakk>n\<rbrakk>) \<Longrightarrow> ((P S)\<lbrakk>n\<rbrakk>) = ((P T)\<lbrakk>n\<rbrakk>)"
using scc_def by blast


text{* The following lemma establishes that @{term scc} is indeed logically stronger as @{term cc}
as its name suggests. *}   
lemma X: "!\<langle>P\<rangle> \<Longrightarrow> \<langle>P\<rangle>"
unfolding scc_def cc_def
by auto (meson equal_upto_def)

lemma scc_const[simp] : "!\<langle>\<lambda>_. c\<rangle>"
by (auto simp: scc_def)

lemma scc_id[simp] :  "!\<langle>\<lambda>X. X\<rangle>"
by (auto simp: scc_def)

lemma cc_const[simp] : "\<langle>\<lambda>_. c\<rangle>"
by (auto simp: cc_def equal_upto_def)

lemma cc_id[simp] :  "\<langle>\<lambda>X. X\<rangle>"
by (auto simp: cc_def)
 

lemma cc_subst2 :  "\<langle>P\<rangle> \<Longrightarrow> \<langle>Q\<rangle> \<Longrightarrow> X\<triangleq>\<^sub>\<lbrakk>\<^bsub>n\<^esub>\<^sub>\<rbrakk>Y \<Longrightarrow> (P Y)\<lbrakk>n\<rbrakk> = (Q Y)\<lbrakk>n\<rbrakk> \<Longrightarrow>  (P X)\<lbrakk>n\<rbrakk> = (Q X)\<lbrakk>n\<rbrakk>"
by (simp add: cc_def equal_upto_def)

lemma scc_subst2 : "!\<langle>P\<rangle> \<Longrightarrow> !\<langle>Q\<rangle> \<Longrightarrow>  X\<lbrakk>n\<rbrakk> = Y\<lbrakk>n\<rbrakk>\<Longrightarrow> (P Y)\<lbrakk>n\<rbrakk> = (Q Y)\<lbrakk>n\<rbrakk> \<Longrightarrow>  (P X)\<lbrakk>n\<rbrakk> = (Q X)\<lbrakk>n\<rbrakk>"
unfolding scc_def
by metis

lemma scc_subst2_rev : "X\<lbrakk>n\<rbrakk> = null \<Longrightarrow> !\<langle>P\<rangle> \<Longrightarrow> !\<langle>Q\<rangle> \<Longrightarrow> (P null)\<lbrakk>n\<rbrakk> = (Q null)\<lbrakk>n\<rbrakk> \<Longrightarrow> (P X)\<lbrakk>n\<rbrakk> = (Q X)\<lbrakk>n\<rbrakk>"
apply(rule scc_subst2[of P Q]) 
by simp_all



subsection{* Constant Value operators *}

definition const_stream ::"('a::null stream) \<Rightarrow> bool"
where     "const_stream S \<equiv> (\<forall> n n'. (S\<lbrakk>n\<rbrakk>) = (S\<lbrakk>n'\<rbrakk>))"  

lemma [simp] : "const_stream null" unfolding const_stream_def by simp



locale bin_scheme =
   fixes f::"('a::null)stream \<Rightarrow> ('b::null)stream \<Rightarrow> ('c::null)stream"
   fixes g::"'a::null \<Rightarrow> 'b::null \<Rightarrow> 'c::null" 
   assumes def_scheme: "\<forall> X Y. f X Y = Abs_stream(\<lambda> n. if  is_def(X\<lbrakk>n\<rbrakk>) \<and> is_def(Y\<lbrakk>n\<rbrakk>) 
                                                       then g (X\<lbrakk>n\<rbrakk>)(Y\<lbrakk>n\<rbrakk>) 
                                                       else null)"
   assumes total_core: "\<And> X Y. is_def X \<Longrightarrow> is_def Y \<Longrightarrow>is_def (g X Y)"
begin

lemma scc_preserve [simp]: "!\<langle>S\<rangle> \<Longrightarrow> !\<langle>T\<rangle> \<Longrightarrow>  !\<langle>\<lambda>X. f (S X) (T X)\<rangle> "
unfolding not_def map_fun_def scc_def equal_upto_def sel_def o_def 
apply(auto simp:def_scheme )
apply(subst Abs_stream_inverse')
apply(subst Abs_stream_inverse')
unfolding sel_def
by (metis )

lemma cc_preserve[simp] : "\<langle>S\<rangle> \<Longrightarrow> \<langle>T\<rangle> \<Longrightarrow> \<langle>\<lambda>X. f (S X) (T X)\<rangle> "
unfolding not_def map_fun_def cc_def equal_upto_def sel_def
by(auto simp:  def_scheme Abs_stream_inverse' sel_def)

lemma cc_split [simp]: "\<langle>\<lambda>\<lbrakk>x, y\<rbrakk>. f x y\<rangle>"
unfolding def_scheme map_fun_def cc_def scc_def sel_def split\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m_def
apply(auto simp: Abs_stream_inverse' o_def cc_def equal_upto_def sel_def def_scheme)
by (simp_all add: fst\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m.rep_eq snd\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m.rep_eq)

lemma scc_split[simp] : "!\<langle>\<lambda>\<lbrakk>x, y\<rbrakk>. f x y\<rangle>"
unfolding not_def map_fun_def scc_def sel_def
by (simp add: Abs_stream_inverse' def_scheme fst\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m.rep_eq sel_def snd\<^sub>s\<^sub>t\<^sub>r\<^sub>e\<^sub>a\<^sub>m.rep_eq)

lemma strict1[simp]:"(f null X) = null" 
unfolding not_def sel_def  null_stream_def nil_def
by(auto simp: Abs_stream_inverse' def_scheme sel_def)

lemma strict2[simp]:"(f X null) = null" 
unfolding not_def sel_def  null_stream_def nil_def
by(auto simp: Abs_stream_inverse' def_scheme sel_def)

lemma strict':"X\<lbrakk>n\<rbrakk> = null \<Longrightarrow> (f X Y)\<lbrakk>n\<rbrakk> = null" 
unfolding not_def map_fun_def o_def sel_def
by(simp add: Abs_stream_inverse'  def_scheme sel_def)

lemma strict'':"Y\<lbrakk>n\<rbrakk> = null \<Longrightarrow> (f X Y)\<lbrakk>n\<rbrakk> = null" 
unfolding not_def map_fun_def o_def sel_def
by(simp add: Abs_stream_inverse'  def_scheme sel_def)


lemma not_def_prop[simp]: "is_def ((f X Y)\<lbrakk>n\<rbrakk>) = (is_def (X\<lbrakk>n\<rbrakk>) \<and> is_def (Y\<lbrakk>n\<rbrakk>))"
unfolding not_def map_fun_def o_def sel_def null_stream_def
by(auto simp: Abs_stream_inverse' def_scheme sel_def total_core)

lemma const_stream_prop [simp]: 
"const_stream X \<Longrightarrow>  const_stream Y \<Longrightarrow> const_stream (f X Y) "
by (smt Abs_stream_inverse' const_stream_def def_scheme sel_def)

(* in a sub_locale for option-based streams we can have:

lemma not_nonstrict: "is_def (S\<lbrakk>n\<rbrakk>) \<Longrightarrow>  is_def (T\<lbrakk>n\<rbrakk>) \<Longrightarrow> (f S T)\<lbrakk>n\<rbrakk> = Some(g (the(S\<lbrakk>n\<rbrakk>) (the(ST\<lbrakk>n\<rbrakk>))"
unfolding not_def map_fun_def o_def sel_def  
by(auto simp: Abs_stream_inverse')

which is crucial for computational reasoning.

*)

end


locale bin_scheme_option =
   fixes f::"('a option)stream \<Rightarrow> ('b option)stream \<Rightarrow> ('c option)stream"
   fixes g::"'a  \<Rightarrow> 'b  \<Rightarrow> 'c " 
   assumes def_scheme: "\<forall> X Y. f X Y = Abs_stream(\<lambda> n. if  is_def(X\<lbrakk>n\<rbrakk>) \<and> is_def(Y\<lbrakk>n\<rbrakk>) 
                                                       then Some(g (the (X\<lbrakk>n\<rbrakk>))(the (Y\<lbrakk>n\<rbrakk>))) 
                                                       else null)"
begin

lemma nonstrict_compute[simp]: 
"is_def (S\<lbrakk>n\<rbrakk>) \<Longrightarrow>  is_def (T\<lbrakk>n\<rbrakk>) \<Longrightarrow> (f S T)\<lbrakk>n\<rbrakk> = Some(g (the (S\<lbrakk>n\<rbrakk>)) (the (T\<lbrakk>n\<rbrakk>)))"
by(auto simp: def_scheme sel_def Abs_stream_inverse') 

(* problem solved: property inheritance from bin_scheme via sub - (locale instantiation) 
   instead of (sub-locale) - instantiation. *) 
interpretation sub : bin_scheme f "\<lambda> x y. Some (g (the x) (the y))"
apply unfold_locales
apply(subst stream_ext)
by(auto simp: def_scheme sel_def Abs_stream_inverse')

lemmas strict1 [simp] = local.sub.strict1 (* f null ?X = null *)
lemmas strict2 [simp]= local.sub.strict2 (* f ?X null = null *)
lemmas cc_split[simp]=  local.sub.cc_split    (* \<langle>\<lambda>\<lbrakk>x, y\<rbrakk>. f x y\<rangle> *)
lemmas scc_split[simp]=  local.sub.scc_split   (* !\<langle>\<lambda>\<lbrakk>x, y\<rbrakk>. f x y\<rangle>*)
lemmas const_stream_prop[simp] =  local.sub.const_stream_prop
                       (*  const_stream ?X \<Longrightarrow> const_stream ?Y \<Longrightarrow> const_stream (f ?X ?Y) *)
lemmas strict'  [simp]    =  local.sub.strict'      (* ?X\<lbrakk>?n\<rbrakk> = null \<Longrightarrow> f ?X ?Y\<lbrakk>?n\<rbrakk> = null     *) 
lemmas strict'' [simp]    =  local.sub.strict''     (* ?Y\<lbrakk>?n\<rbrakk> = null \<Longrightarrow> f ?X ?Y\<lbrakk>?n\<rbrakk> = null     *)
lemmas cc_preserve [simp] =  local.sub.cc_preserve  (* \<langle>?S\<rangle> \<Longrightarrow> \<langle>?T\<rangle> \<Longrightarrow> \<langle>\<lambda>X. f (?S X) (?T X)\<rangle>    *)
lemmas scc_preserve[simp] =  local.sub.scc_preserve (* !\<langle>?S\<rangle> \<Longrightarrow> !\<langle>?T\<rangle> \<Longrightarrow> !\<langle>\<lambda>X. f (?S X) (?T X)\<rangle> *)
lemmas total_core  [simp] =  local.sub.total_core   (* is_def ?X \<Longrightarrow> is_def ?Y \<Longrightarrow> 
                                                 is_def (Some (g (the ?X) (the ?Y))) *)
lemmas not_def_prop [simp] =  local.sub.not_def_prop (* is_def (f ?X ?Y\<lbrakk>?n\<rbrakk>) = 
                                                 (is_def (?X\<lbrakk>?n\<rbrakk>) \<and> is_def (?Y\<lbrakk>?n\<rbrakk>)) *)

end



locale un_scheme =
   fixes f::"('a::null)stream \<Rightarrow> ('b::null)stream"
   fixes g::"'a::null \<Rightarrow> 'b::null" 
   assumes def_scheme: "\<forall> X. f X = Abs_stream(\<lambda> n. if  is_def(X\<lbrakk>n\<rbrakk>) then g (X\<lbrakk>n\<rbrakk>) else null)"
   assumes total_core: "\<And> X. is_def X \<Longrightarrow> is_def (g X)"
begin

lemma scc_preserve [simp]: "!\<langle>S\<rangle> \<Longrightarrow> !\<langle>\<lambda>X. f (S X)\<rangle> "
unfolding not_def map_fun_def scc_def equal_upto_def sel_def o_def 
apply(auto simp:def_scheme )
apply(subst Abs_stream_inverse')
apply(subst Abs_stream_inverse')
unfolding sel_def
by (metis )

lemma cc_preserve[simp] : "\<langle>S\<rangle> \<Longrightarrow> \<langle>\<lambda>X. f (S X)\<rangle> "
unfolding not_def map_fun_def cc_def equal_upto_def sel_def
by(auto simp:  def_scheme Abs_stream_inverse' sel_def)

lemma strict[simp]:"(f null) = null" 
unfolding not_def sel_def  null_stream_def nil_def
by(auto simp: Abs_stream_inverse' def_scheme sel_def)

lemma strict':"X\<lbrakk>n\<rbrakk> = null \<Longrightarrow> (f X)\<lbrakk>n\<rbrakk> = null" 
unfolding not_def map_fun_def o_def sel_def
by(simp add: Abs_stream_inverse'  def_scheme sel_def)

lemma not_def_prop[simp]: "is_def ((f X)\<lbrakk>n\<rbrakk>) = (is_def (X\<lbrakk>n\<rbrakk>))"
unfolding not_def map_fun_def o_def sel_def null_stream_def
by(auto simp: Abs_stream_inverse' def_scheme sel_def total_core)

lemma const_stream_prop [simp]: 
"const_stream X \<Longrightarrow>  const_stream (f X) "
by (smt Abs_stream_inverse' const_stream_def def_scheme sel_def)


end


locale un_scheme_option =
   fixes f::"('a option)stream \<Rightarrow> ('b option)stream"
   fixes g::"'a  \<Rightarrow> 'b" 
   assumes def_scheme: "\<forall> X. f X = Abs_stream(\<lambda> n. if is_def(X\<lbrakk>n\<rbrakk>) then Some(g (the (X\<lbrakk>n\<rbrakk>))) else null)"
begin

lemma nonstrict_compute[simp]: 
"is_def (S\<lbrakk>n\<rbrakk>) \<Longrightarrow>  (f S)\<lbrakk>n\<rbrakk> = Some(g (the (S\<lbrakk>n\<rbrakk>)))"
by(auto simp: def_scheme sel_def Abs_stream_inverse') 

(* problem solved: property inheritance from bin_scheme via sub - (locale instantiation) 
   instead of (sub-locale) - instantiation. *) 
interpretation sub : un_scheme f "\<lambda> x. Some (g (the x))"
apply unfold_locales
apply(subst stream_ext)
by(auto simp: def_scheme sel_def Abs_stream_inverse')

lemmas strict [simp] = local.sub.strict (* f null = null *)
lemmas const_stream_prop[simp] =  local.sub.const_stream_prop (* const_stream ?X \<Longrightarrow> const_stream (f ?X ) *)
lemmas strict'  [simp]    =  local.sub.strict'      (* ?X\<lbrakk>?n\<rbrakk> = null \<Longrightarrow> f ?X\<lbrakk>?n\<rbrakk> = null     *) 
lemmas cc_preserve [simp] =  local.sub.cc_preserve  (* \<langle>?S\<rangle> \<Longrightarrow> \<langle>\<lambda>X. f (?S X)\<rangle>    *)
lemmas scc_preserve[simp] =  local.sub.scc_preserve (* !\<langle>?S\<rangle> \<Longrightarrow> !\<langle>\<lambda>X. f (?S X)\<rangle> *)
lemmas total_core  [simp] =  local.sub.total_core   (* is_def ?X \<Longrightarrow> is_def (Some (g (the ?X))) *)
lemmas not_def_prop [simp] =  local.sub.not_def_prop (* is_def (f ?X\<lbrakk>?n\<rbrakk>) = is_def (?X\<lbrakk>?n\<rbrakk>) *)

end




end 